from datetime import datetime

import numpy as np
import pandas as pd
import pytest
from woodwork.exceptions import TypeConversionError
from woodwork.logical_types import (
    Boolean,
    Categorical,
    Datetime,
    Double,
    Integer,
    NaturalLanguage,
)

from featuretools.entityset.entityset import LTI_COLUMN_NAME, EntitySet
from featuretools.tests.testing_utils import to_pandas
from featuretools.utils.gen_utils import Library, import_or_none, is_instance

dd = import_or_none("dask.dataframe")
ps = import_or_none("pyspark.pandas")


def test_empty_es():
    es = EntitySet("es")
    assert es.id == "es"
    assert es.dataframe_dict == {}
    assert es.relationships == []
    assert es.time_type is None


@pytest.fixture
def pd_df():
    return pd.DataFrame({"id": [0, 1, 2], "category": ["a", "b", "c"]}).astype(
        {"category": "category"},
    )


@pytest.fixture
def dd_df(pd_df):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(pd_df, npartitions=2)


@pytest.fixture
def spark_df(pd_df):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    return ps.from_pandas(pd_df)


@pytest.fixture(params=["pd_df", "dd_df", "spark_df"])
def df(request):
    return request.getfixturevalue(request.param)


def test_init_es_with_dataframe(df):
    es = EntitySet("es", dataframes={"table": (df, "id")})
    assert es.id == "es"
    assert len(es.dataframe_dict) == 1
    assert es["table"] is df

    assert es["table"].ww.schema is not None
    assert isinstance(es["table"].ww.logical_types["id"], Integer)
    assert isinstance(es["table"].ww.logical_types["category"], Categorical)


def test_init_es_with_woodwork_table_same_name(df):
    df.ww.init(index="id", name="table")
    es = EntitySet("es", dataframes={"table": (df,)})

    assert es.id == "es"
    assert len(es.dataframe_dict) == 1
    assert es["table"] is df

    assert es["table"].ww.schema is not None

    assert es["table"].ww.index == "id"
    assert es["table"].ww.time_index is None

    assert isinstance(es["table"].ww.logical_types["id"], Integer)
    assert isinstance(es["table"].ww.logical_types["category"], Categorical)


def test_init_es_with_woodwork_table_diff_name_error(df):
    df.ww.init(index="id", name="table")
    error = "Naming conflict in dataframes dictionary: dictionary key 'diff_name' does not match dataframe name 'table'"
    with pytest.raises(ValueError, match=error):
        EntitySet("es", dataframes={"diff_name": (df,)})


def test_init_es_with_dataframe_and_params(df):
    logical_types = {"id": "NaturalLanguage", "category": NaturalLanguage}
    semantic_tags = {"category": "new_tag"}
    es = EntitySet(
        "es",
        dataframes={"table": (df, "id", None, logical_types, semantic_tags)},
    )

    assert es.id == "es"
    assert len(es.dataframe_dict) == 1
    assert es["table"] is df

    assert es["table"].ww.schema is not None

    assert es["table"].ww.index == "id"
    assert es["table"].ww.time_index is None

    assert isinstance(es["table"].ww.logical_types["id"], NaturalLanguage)
    assert isinstance(es["table"].ww.logical_types["category"], NaturalLanguage)

    assert es["table"].ww.semantic_tags["id"] == {"index"}
    assert es["table"].ww.semantic_tags["category"] == {"new_tag"}


def test_init_es_with_multiple_dataframes(pd_df):
    second_df = pd.DataFrame({"id": [0, 1, 2, 3], "first_table_id": [1, 2, 2, 1]})

    pd_df.ww.init(name="first_table", index="id")

    es = EntitySet(
        "es",
        dataframes={
            "first_table": (pd_df,),
            "second_table": (
                second_df,
                "id",
                None,
                None,
                {"first_table_id": "foreign_key"},
            ),
        },
    )

    assert len(es.dataframe_dict) == 2
    assert es["first_table"].ww.schema is not None
    assert es["second_table"].ww.schema is not None


def test_add_dataframe_to_es(df):
    es1 = EntitySet("es")
    assert es1.dataframe_dict == {}
    es1.add_dataframe(
        df,
        dataframe_name="table",
        index="id",
        semantic_tags={"category": "new_tag"},
    )
    assert len(es1.dataframe_dict) == 1

    copy_df = df.ww.copy()

    es2 = EntitySet("es")
    assert es2.dataframe_dict == {}
    es2.add_dataframe(copy_df)
    assert len(es2.dataframe_dict) == 1

    assert es1["table"].ww == es2["table"].ww


def test_change_es_dataframe_schema(df):
    df.ww.init(index="id", name="table")
    es = EntitySet("es", dataframes={"table": (df,)})

    assert es["table"].ww.index == "id"

    es["table"].ww.set_index("category")
    assert es["table"].ww.index == "category"


def test_init_es_with_relationships(pd_df):
    second_df = pd.DataFrame({"id": [0, 1, 2, 3], "first_table_id": [1, 2, 2, 1]})

    pd_df.ww.init(name="first_table", index="id")
    second_df.ww.init(name="second_table", index="id")

    es = EntitySet(
        "es",
        dataframes={"first_table": (pd_df,), "second_table": (second_df,)},
        relationships=[("first_table", "id", "second_table", "first_table_id")],
    )

    assert len(es.relationships) == 1

    forward_dataframes = [name for name, _ in es.get_forward_dataframes("second_table")]
    assert forward_dataframes[0] == "first_table"

    relationship = es.relationships[0]
    assert "foreign_key" in relationship.child_column.ww.semantic_tags
    assert "index" in relationship.parent_column.ww.semantic_tags


@pytest.fixture
def dates_df():
    return pd.DataFrame(
        {
            "backwards_order": [8, 7, 6, 5, 4, 3, 2, 1, 0],
            "dates_backwards": [
                "2020-09-09",
                "2020-09-08",
                "2020-09-07",
                "2020-09-06",
                "2020-09-05",
                "2020-09-04",
                "2020-09-03",
                "2020-09-02",
                "2020-09-01",
            ],
            "random_order": [7, 6, 8, 0, 2, 4, 3, 1, 5],
            "repeating_dates": [
                "2020-08-01",
                "2019-08-01",
                "2020-08-01",
                "2012-08-01",
                "2019-08-01",
                "2019-08-01",
                "2019-08-01",
                "2013-08-01",
                "2019-08-01",
            ],
            "special": [7, 8, 0, 1, 4, 2, 6, 3, 5],
            "special_dates": [
                "2020-08-01",
                "2019-08-01",
                "2020-08-01",
                "2012-08-01",
                "2019-08-01",
                "2019-08-01",
                "2019-08-01",
                "2013-08-01",
                "2019-08-01",
            ],
        },
    )


def test_add_secondary_time_index(dates_df):
    dates_df.ww.init(
        name="dates_table",
        index="backwards_order",
        time_index="dates_backwards",
    )
    es = EntitySet("es")
    es.add_dataframe(
        dates_df,
        secondary_time_index={"repeating_dates": ["random_order", "special"]},
    )

    assert dates_df.ww.metadata["secondary_time_index"] == {
        "repeating_dates": ["random_order", "special", "repeating_dates"],
    }


def test_time_type_check_order(dates_df):
    dates_df.ww.init(
        name="dates_table",
        index="backwards_order",
        time_index="random_order",
    )
    es = EntitySet("es")

    error = "dates_table time index is Datetime type which differs from other entityset time indexes"
    with pytest.raises(TypeError, match=error):
        es.add_dataframe(
            dates_df,
            secondary_time_index={"repeating_dates": ["random_order", "special"]},
        )

    assert "secondary_time_index" not in dates_df.ww.metadata


def test_add_time_index_through_woodwork_different_type(dates_df):
    dates_df.ww.init(
        name="dates_table",
        index="backwards_order",
        time_index="dates_backwards",
    )
    es = EntitySet("es")

    es.add_dataframe(
        dates_df,
        secondary_time_index={"repeating_dates": ["random_order", "special"]},
    )

    assert dates_df.ww.metadata["secondary_time_index"] == {
        "repeating_dates": ["random_order", "special", "repeating_dates"],
    }
    assert es.time_type == Datetime

    assert es._check_uniform_time_index(es["dates_table"]) is None

    dates_df.ww.set_time_index("random_order")
    assert dates_df.ww.time_index == "random_order"

    error = "dates_table time index is numeric type which differs from other entityset time indexes"
    with pytest.raises(TypeError, match=error):
        es._check_uniform_time_index(es["dates_table"])


def test_init_with_mismatched_time_types(dates_df):
    dates_df.ww.init(
        name="dates_table",
        index="backwards_order",
        time_index="repeating_dates",
    )
    es = EntitySet("es")
    es.add_dataframe(dates_df, secondary_time_index={"special_dates": ["special"]})
    assert es.time_type == Datetime

    nums_df = pd.DataFrame({"id": [1, 2, 3], "times": [9, 8, 7]})
    nums_df.ww.init(name="numerics_table", index="id", time_index="times")

    error = "numerics_table time index is numeric type which differs from other entityset time indexes"
    with pytest.raises(TypeError, match=error):
        es.add_dataframe(nums_df)


def test_int_double_time_type(dates_df):
    dates_df.ww.init(
        name="dates_table",
        index="backwards_order",
        time_index="random_order",
        logical_types={"random_order": "Integer", "special": "Double"},
    )
    es = EntitySet("es")

    # Both random_order and special are numeric, but they are different logical types
    es.add_dataframe(dates_df, secondary_time_index={"special": ["dates_backwards"]})

    assert isinstance(es["dates_table"].ww.logical_types["random_order"], Integer)
    assert isinstance(es["dates_table"].ww.logical_types["special"], Double)

    assert es["dates_table"].ww.time_index == "random_order"
    assert "special" in es["dates_table"].ww.metadata["secondary_time_index"]


def test_normalize_dataframe():
    df = pd.DataFrame(
        {
            "id": range(4),
            "full_name": [
                "Mr. John Doe",
                "Doe, Mrs. Jane",
                "James Brown",
                "Ms. Paige Turner",
            ],
            "email": [
                "john.smith@example.com",
                np.nan,
                "team@featuretools.com",
                "junk@example.com",
            ],
            "phone_number": [
                "5555555555",
                "555-555-5555",
                "1-(555)-555-5555",
                "555-555-5555",
            ],
            "age": pd.Series([33, None, 33, 57], dtype="Int64"),
            "signup_date": [pd.to_datetime("2020-09-01")] * 4,
            "is_registered": pd.Series([True, False, True, None], dtype="boolean"),
        },
    )

    df.ww.init(name="first_table", index="id", time_index="signup_date")
    es = EntitySet("es")
    es.add_dataframe(df)
    es.normalize_dataframe(
        "first_table",
        "second_table",
        "age",
        additional_columns=["phone_number", "full_name"],
        make_time_index=True,
    )
    assert len(es.dataframe_dict) == 2
    assert "foreign_key" in es["first_table"].ww.semantic_tags["age"]


def test_replace_dataframe():
    df = pd.DataFrame(
        {
            "id": range(4),
            "full_name": [
                "Mr. John Doe",
                "Doe, Mrs. Jane",
                "James Brown",
                "Ms. Paige Turner",
            ],
            "email": [
                "john.smith@example.com",
                np.nan,
                "team@featuretools.com",
                "junk@example.com",
            ],
            "phone_number": [
                "5555555555",
                "555-555-5555",
                "1-(555)-555-5555",
                "555-555-5555",
            ],
            "age": pd.Series([33, None, 33, 57], dtype="Int64"),
            "signup_date": [pd.to_datetime("2020-09-01")] * 4,
            "is_registered": pd.Series([True, False, True, None], dtype="boolean"),
        },
    )

    df.ww.init(name="table", index="id")
    es = EntitySet("es")
    es.add_dataframe(df)
    original_schema = es["table"].ww.schema

    new_df = df.iloc[2:]
    es.replace_dataframe("table", new_df)

    assert len(es["table"]) == 2
    assert es["table"].ww.schema == original_schema


def test_add_last_time_index(es):
    es.add_last_time_indexes(["products"])

    assert "last_time_index" in es["products"].ww.metadata

    assert es["products"].ww.metadata["last_time_index"] == LTI_COLUMN_NAME
    assert LTI_COLUMN_NAME in es["products"]
    assert "last_time_index" in es["products"].ww.semantic_tags[LTI_COLUMN_NAME]
    assert isinstance(es["products"].ww.logical_types[LTI_COLUMN_NAME], Datetime)


def test_add_last_time_non_numeric_index(pd_es, spark_es, dask_es):
    # Confirm that add_last_time_index works for indices that aren't numeric
    # since numeric underlying indices can accidentally match the Woodwork index
    pd_es.add_last_time_indexes(["products"])
    dask_es.add_last_time_indexes(["products"])
    spark_es.add_last_time_indexes(["products"])

    assert list(to_pandas(pd_es["products"][LTI_COLUMN_NAME]).sort_index()) == list(
        to_pandas(dask_es["products"][LTI_COLUMN_NAME]).sort_index(),
    )
    assert list(to_pandas(pd_es["products"][LTI_COLUMN_NAME]).sort_index()) == list(
        to_pandas(spark_es["products"]).sort_values("id")[LTI_COLUMN_NAME],
    )

    assert pd_es["products"].ww.schema == dask_es["products"].ww.schema
    assert pd_es["products"].ww.schema == spark_es["products"].ww.schema


def test_lti_already_has_last_time_column_name(es):
    col = es["customers"].ww.pop("loves_ice_cream")
    col.name = LTI_COLUMN_NAME

    es["customers"].ww[LTI_COLUMN_NAME] = col

    assert LTI_COLUMN_NAME in es["customers"].columns
    assert isinstance(es["customers"].ww.logical_types[LTI_COLUMN_NAME], Boolean)

    error = (
        "Cannot add a last time index on DataFrame with an existing "
        f"'{LTI_COLUMN_NAME}' column. Please rename '{LTI_COLUMN_NAME}'."
    )
    with pytest.raises(ValueError, match=error):
        es.add_last_time_indexes(["customers"])


def test_numeric_es_last_time_index_logical_type(int_es):
    assert int_es.time_type == "numeric"

    int_es.add_last_time_indexes()

    for df in int_es.dataframes:
        assert isinstance(df.ww.logical_types[LTI_COLUMN_NAME], Double)
        int_es._check_uniform_time_index(df, LTI_COLUMN_NAME)


def test_datetime_es_last_time_index_logical_type(es):
    assert es.time_type == Datetime

    es.add_last_time_indexes()

    for df in es.dataframes:
        assert isinstance(df.ww.logical_types[LTI_COLUMN_NAME], Datetime)
        es._check_uniform_time_index(df, LTI_COLUMN_NAME)


def test_dataframe_without_name(es):
    new_es = EntitySet()

    new_df = es["sessions"].copy()

    assert new_df.ww.schema is None

    error = "Cannot add dataframe to EntitySet without a name. Please provide a value for the dataframe_name parameter."
    with pytest.raises(ValueError, match=error):
        new_es.add_dataframe(new_df)


def test_dataframe_with_name_parameter(es):
    new_es = EntitySet()

    new_df = es["sessions"][["id"]]

    assert new_df.ww.schema is None

    new_es.add_dataframe(
        new_df,
        dataframe_name="df_name",
        index="id",
        logical_types={"id": "Integer"},
    )
    assert new_es["df_name"].ww.name == "df_name"


def test_woodwork_dataframe_without_name_errors(es):
    new_es = EntitySet()

    new_df = es["sessions"].ww.copy()
    new_df.ww._schema.name = None

    assert new_df.ww.name is None

    error = "Cannot add a Woodwork DataFrame to EntitySet without a name"
    with pytest.raises(ValueError, match=error):
        new_es.add_dataframe(new_df)


def test_woodwork_dataframe_with_name(es):
    new_es = EntitySet()

    new_df = es["sessions"].ww.copy()
    new_df.ww._schema.name = "df_name"

    assert new_df.ww.name == "df_name"

    new_es.add_dataframe(new_df)

    assert new_es["df_name"].ww.name == "df_name"


def test_woodwork_dataframe_ignore_conflicting_name_parameter_warning(es):
    new_es = EntitySet()

    new_df = es["sessions"].ww.copy()
    new_df.ww._schema.name = "df_name"

    assert new_df.ww.name == "df_name"

    warning = "A Woodwork-initialized DataFrame was provided, so the following parameters were ignored: dataframe_name"
    with pytest.warns(UserWarning, match=warning):
        new_es.add_dataframe(new_df, dataframe_name="conflicting_name")

    assert new_es["df_name"].ww.name == "df_name"


def test_woodwork_dataframe_same_name_parameter(es):
    new_es = EntitySet()

    new_df = es["sessions"].ww.copy()
    new_df.ww._schema.name = "df_name"

    assert new_df.ww.name == "df_name"

    new_es.add_dataframe(new_df, dataframe_name="df_name")

    assert new_es["df_name"].ww.name == "df_name"


def test_extra_woodwork_params(es):
    new_es = EntitySet()

    sessions_df = es["sessions"].ww.copy()

    assert sessions_df.ww.index == "id"
    assert sessions_df.ww.time_index is None
    assert isinstance(sessions_df.ww.logical_types["id"], Integer)

    warning_msg = (
        "A Woodwork-initialized DataFrame was provided, so the following parameters were ignored: "
        "index, time_index, logical_types, make_index, semantic_tags, already_sorted"
    )
    with pytest.warns(UserWarning, match=warning_msg):
        new_es.add_dataframe(
            dataframe_name="sessions",
            dataframe=sessions_df,
            index="filepath",
            time_index="customer_id",
            logical_types={"id": Categorical},
            make_index=True,
            already_sorted=True,
            semantic_tags={"id": "new_tag"},
        )
    assert sessions_df.ww.index == "id"
    assert sessions_df.ww.time_index is None
    assert isinstance(sessions_df.ww.logical_types["id"], Integer)
    assert "new_tag" not in sessions_df.ww.semantic_tags


def test_replace_dataframe_errors(es):
    df = es["customers"].copy()
    if ps and isinstance(df, ps.DataFrame):
        df["new"] = [1, 2, 3]
    else:
        df["new"] = pd.Series([1, 2, 3])

    error_text = "New dataframe is missing new cohort column"
    with pytest.raises(ValueError, match=error_text):
        es.replace_dataframe(dataframe_name="customers", df=df.drop(columns=["cohort"]))

    error_text = "New dataframe contains 16 columns, expecting 15"
    with pytest.raises(ValueError, match=error_text):
        es.replace_dataframe(dataframe_name="customers", df=df)


def test_replace_dataframe_already_sorted(es):
    # test already_sorted on dataframe without time index
    df = es["sessions"].copy()
    updated_id = to_pandas(df["id"])
    updated_id.iloc[1] = 2
    updated_id.iloc[2] = 1

    df = df.set_index("id", drop=False)
    df.index.name = None

    assert es["sessions"].ww.time_index is None

    if ps and isinstance(df, ps.DataFrame):
        df["id"] = updated_id.to_list()
        df = df.sort_index()
    elif is_instance(df, dd, "DataFrame"):
        df["id"] = updated_id

    es.replace_dataframe(dataframe_name="sessions", df=df.copy(), already_sorted=False)
    sessions_df = to_pandas(es["sessions"])
    assert sessions_df["id"].iloc[1] == 2  # no sorting since time index not defined
    es.replace_dataframe(dataframe_name="sessions", df=df.copy(), already_sorted=True)
    sessions_df = to_pandas(es["sessions"])
    assert sessions_df["id"].iloc[1] == 2

    # test already_sorted on dataframe with time index
    df = es["customers"].copy()
    updated_signup = to_pandas(df["signup_date"])
    updated_signup.iloc[0] = datetime(2011, 4, 11)

    assert es["customers"].ww.time_index == "signup_date"

    if ps and isinstance(df, ps.DataFrame):
        df["signup_date"] = updated_signup.to_list()
        df = df.sort_index()
    else:
        df["signup_date"] = updated_signup

    es.replace_dataframe(dataframe_name="customers", df=df.copy(), already_sorted=True)
    customers_df = to_pandas(es["customers"])
    assert customers_df["id"].iloc[0] == 2

    # only pandas allows for sorting:
    es.replace_dataframe(dataframe_name="customers", df=df.copy(), already_sorted=False)
    updated_customers = to_pandas(es["customers"])
    if isinstance(df, pd.DataFrame):
        assert updated_customers["id"].iloc[0] == 0
    else:
        assert updated_customers["id"].iloc[0] == 2


def test_replace_dataframe_invalid_schema(es):
    if es.dataframe_type != Library.PANDAS:
        pytest.xfail(
            "Invalid schema checks able to be caught by Woodwork only relevant for Pandas",
        )
    df = es["customers"].copy()
    df["id"] = pd.Series([1, 1, 1])

    error_text = "Index column must be unique"
    with pytest.raises(IndexError, match=error_text):
        es.replace_dataframe(dataframe_name="customers", df=df)


def test_replace_dataframe_mismatched_index(es):
    if es.dataframe_type != Library.PANDAS:
        pytest.xfail(
            "Only pandas checks whether underlying index matches the Woodwork index",
        )
    df = es["customers"].copy()
    df["id"] = pd.Series([99, 88, 77])

    es.replace_dataframe(dataframe_name="customers", df=df)

    assert all([77, 99, 88] == es["customers"]["id"])
    assert all([77, 99, 88] == (es["customers"]["id"]).index)


def test_replace_dataframe_different_dtypes(es):
    float_dtype_df = es["customers"].copy()
    float_dtype_df = float_dtype_df.astype({"age": "float64"})

    es.replace_dataframe(dataframe_name="customers", df=float_dtype_df)

    assert es["customers"]["age"].dtype == "int64"
    assert isinstance(es["customers"].ww.logical_types["age"], Integer)

    incompatible_dtype_df = es["customers"].copy()
    incompatible_list = ["hi", "bye", "bye"]
    if ps and isinstance(incompatible_dtype_df, ps.DataFrame):
        incompatible_dtype_df["age"] = incompatible_list
    else:
        incompatible_dtype_df["age"] = pd.Series(incompatible_list)

    if isinstance(es["customers"], pd.DataFrame):
        # Dask and Spark do not error on invalid type conversion until compute
        error_msg = "Error converting datatype for age from type object to type int64. Please confirm the underlying data is consistent with logical type Integer."
        with pytest.raises(TypeConversionError, match=error_msg):
            es.replace_dataframe(dataframe_name="customers", df=incompatible_dtype_df)


@pytest.fixture()
def latlong_df_pandas():
    latlong_df = pd.DataFrame(
        {
            "tuples": pd.Series([(1, 2), (3, 4)]),
            "string_tuple": pd.Series(["(1, 2)", "(3, 4)"]),
            "bracketless_string_tuple": pd.Series(["1, 2", "3, 4"]),
            "list_strings": pd.Series([["1", "2"], ["3", "4"]]),
            "combo_tuple_types": pd.Series(["[1, 2]", "(3, 4)"]),
        },
    )
    latlong_df.set_index("string_tuple", drop=False, inplace=True)
    latlong_df.index.name = None
    return latlong_df


@pytest.fixture()
def latlong_df_dask(latlong_df_pandas):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(latlong_df_pandas, npartitions=2)


@pytest.fixture()
def latlong_df_spark(latlong_df_pandas):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    return ps.from_pandas(
        latlong_df_pandas.applymap(
            lambda tup: list(tup) if isinstance(tup, tuple) else tup,
        ),
    )


@pytest.fixture(params=["latlong_df_pandas", "latlong_df_dask", "latlong_df_spark"])
def latlong_df(request):
    return request.getfixturevalue(request.param)


def test_replace_dataframe_data_transformation(latlong_df):
    initial_df = latlong_df.copy()
    initial_df.ww.init(
        name="latlongs",
        index="string_tuple",
        logical_types={col_name: "LatLong" for col_name in initial_df.columns},
    )
    es = EntitySet()
    es.add_dataframe(dataframe=initial_df)

    df = to_pandas(es["latlongs"])
    expected_val = (1, 2)
    if ps and isinstance(es["latlongs"], ps.DataFrame):
        expected_val = [1, 2]
    for col in latlong_df.columns:
        series = df[col]
        assert series.iloc[0] == expected_val

    es.replace_dataframe("latlongs", latlong_df)
    df = to_pandas(es["latlongs"])
    expected_val = (3, 4)
    if ps and isinstance(es["latlongs"], ps.DataFrame):
        expected_val = [3, 4]
    for col in latlong_df.columns:
        series = df[col]
        assert series.iloc[-1] == expected_val


def test_replace_dataframe_column_order(es):
    original_column_order = es["customers"].columns.copy()

    df = es["customers"].copy()
    col = df.pop("cohort")
    df[col.name] = col

    assert not df.columns.equals(original_column_order)
    assert set(df.columns) == set(original_column_order)

    es.replace_dataframe(dataframe_name="customers", df=df)

    assert es["customers"].columns.equals(original_column_order)


def test_replace_dataframe_different_woodwork_initialized(es):
    df = es["customers"].copy()
    if ps and isinstance(df, ps.DataFrame):
        df["age"] = [1, 2, 3]
    else:
        df["age"] = pd.Series([1, 2, 3])

    # Initialize Woodwork on the new DataFrame and change the schema so it won't match the original DataFrame's schema
    df.ww.init(schema=es["customers"].ww.schema)
    df.ww.set_types(
        logical_types={"id": "NaturalLanguage", "cancel_date": "NaturalLanguage"},
    )
    assert df["id"].dtype == "string"
    assert df["cancel_date"].dtype == "string"

    assert es["customers"]["id"].dtype == "int64"
    assert es["customers"]["cancel_date"].dtype == "datetime64[ns]"

    original_schema = es["customers"].ww.schema

    warning = "Woodwork typing information on new dataframe will be replaced with existing typing information from customers"
    with pytest.warns(UserWarning, match=warning):
        es.replace_dataframe("customers", df, already_sorted=True)

    actual = to_pandas(es["customers"]["age"]).sort_values()
    assert all(actual == [1, 2, 3])

    assert es["customers"].ww._schema == original_schema
    assert es["customers"]["id"].dtype == "int64"
    assert es["customers"]["cancel_date"].dtype == "datetime64[ns]"


@pytest.mark.skipif("not dd")
def test_replace_dataframe_different_dataframe_types():
    dask_es = EntitySet(id="dask_es")

    sessions = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "user": [1, 2, 1, 3],
            "time": [
                pd.to_datetime("2019-01-10"),
                pd.to_datetime("2019-02-03"),
                pd.to_datetime("2019-01-01"),
                pd.to_datetime("2017-08-25"),
            ],
            "strings": ["I am a string", "23", "abcdef ghijk", ""],
        },
    )
    sessions_dask = dd.from_pandas(sessions, npartitions=2)
    sessions_logical_types = {
        "id": Integer,
        "user": Integer,
        "time": Datetime,
        "strings": NaturalLanguage,
    }
    sessions_semantic_tags = {"user": "foreign_key"}

    dask_es.add_dataframe(
        dataframe_name="sessions",
        dataframe=sessions_dask,
        index="id",
        time_index="time",
        logical_types=sessions_logical_types,
        semantic_tags=sessions_semantic_tags,
    )

    with pytest.raises(TypeError, match="Incorrect DataFrame type used"):
        dask_es.replace_dataframe("sessions", sessions)


def test_replace_dataframe_and_min_last_time_index(es):
    es.add_last_time_indexes(["products"])

    original_time_index = es["log"]["datetime"].copy()
    original_last_time_index = es["products"][LTI_COLUMN_NAME].copy()

    if ps and isinstance(original_time_index, ps.Series):
        new_time_index = ps.from_pandas(
            original_time_index.to_pandas() + pd.Timedelta(days=1),
        )
        expected_last_time_index = ps.from_pandas(
            original_last_time_index.to_pandas() + pd.Timedelta(days=1),
        )
    else:
        new_time_index = original_time_index + pd.Timedelta(days=1)
        expected_last_time_index = original_last_time_index + pd.Timedelta(days=1)

    new_dataframe = es["log"].copy()
    new_dataframe["datetime"] = new_time_index
    new_dataframe.pop(LTI_COLUMN_NAME)

    es.replace_dataframe("log", new_dataframe, recalculate_last_time_indexes=True)

    # Spark reorders indices during last time index, so we sort to confirm individual values are the same
    pd.testing.assert_series_equal(
        to_pandas(es["products"][LTI_COLUMN_NAME]).sort_index(),
        to_pandas(expected_last_time_index).sort_index(),
    )
    pd.testing.assert_series_equal(
        to_pandas(es["log"][LTI_COLUMN_NAME]).sort_index(),
        to_pandas(new_time_index).sort_index(),
        check_names=False,
    )


def test_replace_dataframe_dont_recalculate_last_time_index_present(es):
    es.add_last_time_indexes()

    original_time_index = es["customers"]["signup_date"].copy()
    original_last_time_index = es["customers"][LTI_COLUMN_NAME].copy()

    if ps and isinstance(original_time_index, ps.Series):
        new_time_index = ps.from_pandas(
            original_time_index.to_pandas() + pd.Timedelta(days=10),
        )
    else:
        new_time_index = original_time_index + pd.Timedelta(days=10)

    new_dataframe = es["customers"].copy()
    new_dataframe["signup_date"] = new_time_index

    es.replace_dataframe(
        "customers",
        new_dataframe,
        recalculate_last_time_indexes=False,
    )
    pd.testing.assert_series_equal(
        to_pandas(es["customers"][LTI_COLUMN_NAME], sort_index=True),
        to_pandas(original_last_time_index, sort_index=True),
    )


def test_replace_dataframe_dont_recalculate_last_time_index_not_present(es):
    es.add_last_time_indexes()
    original_lti_name = es["customers"].ww.metadata.get("last_time_index")
    assert original_lti_name is not None

    original_time_index = es["customers"]["signup_date"].copy()

    if ps and isinstance(original_time_index, ps.Series):
        new_time_index = ps.from_pandas(
            original_time_index.to_pandas() + pd.Timedelta(days=10),
        )
    else:
        new_time_index = original_time_index + pd.Timedelta(days=10)

    new_dataframe = es["customers"].copy()
    new_dataframe["signup_date"] = new_time_index
    new_dataframe.pop(LTI_COLUMN_NAME)

    es.replace_dataframe(
        "customers",
        new_dataframe,
        recalculate_last_time_indexes=False,
    )
    assert "last_time_index" not in es["customers"].ww.metadata
    assert original_lti_name not in es["customers"].columns


def test_replace_dataframe_recalculate_last_time_index_not_present(es):
    es.add_last_time_indexes()

    original_time_index = es["log"]["datetime"].copy()

    if ps and isinstance(original_time_index, ps.Series):
        new_time_index = ps.from_pandas(
            original_time_index.to_pandas() + pd.Timedelta(days=10),
        )
    else:
        new_time_index = original_time_index + pd.Timedelta(days=10)

    new_dataframe = es["log"].copy()
    new_dataframe["datetime"] = new_time_index
    new_dataframe.pop(LTI_COLUMN_NAME)

    es.replace_dataframe("log", new_dataframe, recalculate_last_time_indexes=True)
    pd.testing.assert_series_equal(
        to_pandas(es["log"]["datetime"]).sort_index(),
        to_pandas(new_time_index).sort_index(),
        check_names=False,
    )
    pd.testing.assert_series_equal(
        to_pandas(es["log"][LTI_COLUMN_NAME]).sort_index(),
        to_pandas(new_time_index).sort_index(),
        check_names=False,
    )


def test_replace_dataframe_recalculate_last_time_index_present(es):
    es.add_last_time_indexes()

    original_time_index = es["log"]["datetime"].copy()

    if ps and isinstance(original_time_index, ps.Series):
        new_time_index = ps.from_pandas(
            original_time_index.to_pandas() + pd.Timedelta(days=10),
        )
    else:
        new_time_index = original_time_index + pd.Timedelta(days=10)

    new_dataframe = es["log"].copy()
    new_dataframe["datetime"] = new_time_index
    assert LTI_COLUMN_NAME in new_dataframe.columns

    es.replace_dataframe("log", new_dataframe, recalculate_last_time_indexes=True)
    pd.testing.assert_series_equal(
        to_pandas(es["log"]["datetime"]).sort_index(),
        to_pandas(new_time_index).sort_index(),
        check_names=False,
    )
    pd.testing.assert_series_equal(
        to_pandas(es["log"][LTI_COLUMN_NAME]).sort_index(),
        to_pandas(new_time_index).sort_index(),
        check_names=False,
    )


def test_normalize_dataframe_loses_column_metadata(es):
    es["log"].ww.columns["value"].metadata["interesting_values"] = [0.0, 1.0]
    es["log"].ww.columns["priority_level"].metadata["interesting_values"] = [1]

    es["log"].ww.columns["value"].description = "a value column"
    es["log"].ww.columns["priority_level"].description = "a priority level column"

    assert "interesting_values" in es["log"].ww.columns["priority_level"].metadata
    assert "interesting_values" in es["log"].ww.columns["value"].metadata
    assert es["log"].ww.columns["value"].description == "a value column"
    assert (
        es["log"].ww.columns["priority_level"].description == "a priority level column"
    )

    es.normalize_dataframe(
        "log",
        "values_2",
        "value_2",
        additional_columns=["priority_level"],
        copy_columns=["value"],
        make_time_index=False,
    )

    # Metadata in the original dataframe and the new dataframe are maintained
    assert "interesting_values" in es["log"].ww.columns["value"].metadata
    assert "interesting_values" in es["values_2"].ww.columns["value"].metadata
    assert "interesting_values" in es["values_2"].ww.columns["priority_level"].metadata
    assert es["log"].ww.columns["value"].description == "a value column"
    assert es["values_2"].ww.columns["value"].description == "a value column"
    assert (
        es["values_2"].ww.columns["priority_level"].description
        == "a priority level column"
    )


def test_normalize_ww_init():
    es = EntitySet()
    df = pd.DataFrame(
        {
            "id": [1, 2, 3, 4],
            "col": ["a", "b", "c", "d"],
            "df2_id": [1, 1, 2, 2],
            "df2_col": [True, False, True, True],
        },
    )

    df.ww.init(index="id", name="test_name")
    es.add_dataframe(dataframe=df)

    assert es["test_name"].ww.name == "test_name"
    assert es["test_name"].ww.schema.name == "test_name"

    es.normalize_dataframe(
        "test_name",
        "new_df",
        "df2_id",
        additional_columns=["df2_col"],
    )

    assert es["test_name"].ww.name == "test_name"
    assert es["test_name"].ww.schema.name == "test_name"

    assert es["new_df"].ww.name == "new_df"
    assert es["new_df"].ww.schema.name == "new_df"



from __future__ import annotations

from typing import Dict, List

import pandas as pd
from woodwork.logical_types import LogicalType

from featuretools.feature_base.feature_base import (
    FeatureBase,
    IdentityFeature,
    TransformFeature,
)
from featuretools.feature_discovery.LiteFeature import LiteFeature
from featuretools.primitives import TransformPrimitive
from featuretools.primitives.base.primitive_base import PrimitiveBase

FeatureCache = Dict[str, FeatureBase]


def convert_featurebase_list_to_feature_list(
    featurebase_list: List[FeatureBase],
) -> List[LiteFeature]:
    """
    Convert a List of FeatureBase objects to a list LiteFeature objects

    Args:
        featurebase_list (List[FeatureBase]):

    Returns:
       LiteFeatures (List[LiteFeature]) - converted LiteFeature objects
    """

    def rfunc(fb: FeatureBase) -> List[LiteFeature]:
        base_features = [
            feature
            for feature_list in [rfunc(x) for x in fb.base_features]
            for feature in feature_list
        ]
        col_schema = fb.column_schema

        logical_type = col_schema.logical_type
        if logical_type is not None:
            assert issubclass(type(logical_type), LogicalType)
            logical_type = type(logical_type)

        tags = col_schema.semantic_tags

        if isinstance(fb, IdentityFeature):
            primitive = None
        else:
            primitive = fb.primitive
            assert isinstance(primitive, PrimitiveBase)

        if fb.number_output_features > 1:
            features: List[LiteFeature] = []

            for idx, name in enumerate(fb.get_feature_names()):
                f = LiteFeature(
                    name=name,
                    logical_type=logical_type,
                    tags=tags,
                    primitive=primitive,
                    base_features=base_features,
                    # TODO: use when working with multi-table
                    df_id=None,
                    idx=idx,
                )
                features.append(f)

            for feature in features:
                related_features = [f for f in features if f.id != feature.id]
                feature.related_features = set(related_features)

            return features

        return [
            LiteFeature(
                name=fb.get_name(),
                logical_type=logical_type,
                tags=tags,
                primitive=primitive,
                base_features=base_features,
                # TODO: use when working with multi-table
                df_id=None,
            ),
        ]

    return [
        feature
        for feature_list in [rfunc(fb) for fb in featurebase_list]
        for feature in feature_list
    ]


def _feature_to_transform_feature(
    feature: LiteFeature,
    base_features: List[FeatureBase],
) -> FeatureBase:
    """
    Transform LiteFeature into FeatureBase object. Handles the Multi-output
    feature in correct way.

    Args:
        feature (LiteFeature)
        base_features (List[FeatureBase])

    Returns:
       FeatureBase
    """
    assert feature.primitive

    assert isinstance(
        feature.primitive,
        TransformPrimitive,
    ), "Only Transform Primitives"

    fb = TransformFeature(base_features, feature.primitive)
    if feature.is_multioutput():
        sorted_features = sorted(
            [f for f in feature.related_features] + [feature],
            key=lambda x: x.idx,
        )
        names = [x.name for x in sorted_features]

        fb = fb.rename(feature.non_indexed_name)
        fb.set_feature_names(names)
    else:
        fb = fb.rename(feature.name)

    return fb


def _convert_feature_to_featurebase(
    feature: LiteFeature,
    dataframe: pd.DataFrame,
    cache: FeatureCache,
) -> FeatureBase:
    """
    Recursively transforms a LiteFeature object into a Featurebase object

    Args:
        feature (LiteFeature)
        base_features (List[FeatureBase])
        cache (FeatureCache) already converted features

    Returns:
       FeatureBase
    """

    def get_base_features(
        feature: LiteFeature,
    ) -> List[FeatureBase]:
        new_base_features: List[FeatureBase] = []
        for bf in feature.base_features:
            fb = rfunc(bf)
            if bf.is_multioutput():
                idx = bf.idx
                # if its multioutput, you can index on the FeatureBase
                new_base_features.append(fb[idx])
            else:
                new_base_features.append(fb)

        return new_base_features

    def rfunc(feature: LiteFeature) -> FeatureBase:
        # if feature has already been converted, return from cache
        if feature.id in cache:
            return cache[feature.id]

        # if depth is 0, we are at an origin feature
        if feature.depth == 0:
            fb = IdentityFeature(dataframe.ww[feature.name])
            cache[feature.id] = fb
            return fb

        base_features = get_base_features(feature)

        fb = _feature_to_transform_feature(feature, base_features)
        cache[feature.id] = fb
        return fb

    return rfunc(feature)


def convert_feature_list_to_featurebase_list(
    feature_list: List[LiteFeature],
    dataframe: pd.DataFrame,
) -> List[FeatureBase]:
    """
    Convert a list of LiteFeature objects into a list of FeatureBase objects

    Args:
        feature_list (List[LiteFeature])
        dataframe (pd.DataFrame)

    Returns:
       List[FeatureBase]
    """
    feature_cache: FeatureCache = {}

    converted_features: List[FeatureBase] = []
    for feature in feature_list:
        if feature.is_multioutput():
            related_feature_ids = [f.id for f in feature.related_features]
            if any((x in feature_cache for x in related_feature_ids)):
                # feature base already created for related ids
                continue

        fb = _convert_feature_to_featurebase(
            feature=feature,
            dataframe=dataframe,
            cache=feature_cache,
        )
        converted_features.append(fb)

    return converted_features



import numpy as np
import pandas as pd
import pytest

from featuretools.primitives import AbsoluteDiff


class TestAbsoluteDiff:
    def test_nan(self):
        data = pd.Series([np.nan, 5, 10, 20, np.nan, 10, np.nan])
        answer = pd.Series([np.nan, np.nan, 5, 10, 0, 10, 0])
        primitive_func = AbsoluteDiff().get_function()
        given_answer = primitive_func(data)
        np.testing.assert_array_equal(given_answer, answer)

    def test_regular(self):
        data = pd.Series([2, 5, 15, 3, 9, 4.5])
        answer = pd.Series([np.nan, 3, 10, 12, 6, 4.5])
        primitive_func = AbsoluteDiff().get_function()
        given_answer = primitive_func(data)
        np.testing.assert_array_equal(given_answer, answer)

    def test_method(self):
        data = pd.Series([2, np.nan, 15, 3, np.nan, 4.5])
        answer = pd.Series([np.nan, 13, 0, 12, 1.5, 0])
        primitive_func = AbsoluteDiff(method="backfill").get_function()
        given_answer = primitive_func(data)
        np.testing.assert_array_equal(given_answer, answer)

    def test_limit(self):
        data = pd.Series([2, np.nan, np.nan, np.nan, 3.0, 4.5])
        answer = pd.Series([np.nan, 0, 0, np.nan, np.nan, 1.5])
        primitive_func = AbsoluteDiff(limit=2).get_function()
        given_answer = primitive_func(data)
        np.testing.assert_array_equal(given_answer, answer)

    def test_zero(self):
        data = pd.Series([2, 0, 0, 5, 0, -4])
        answer = pd.Series([np.nan, 2, 0, 5, 5, 4])
        primitive_func = AbsoluteDiff().get_function()
        given_answer = primitive_func(data)
        np.testing.assert_array_equal(given_answer, answer)

    def test_empty(self):
        data = pd.Series([], dtype="float64")
        answer = pd.Series([], dtype="float64")
        primitive_func = AbsoluteDiff().get_function()
        given_answer = primitive_func(data)
        np.testing.assert_array_equal(given_answer, answer)

    def test_inf(self):
        data = pd.Series([0, np.inf, 0, 5, np.NINF, np.inf, np.NINF])
        answer = pd.Series([np.nan, np.inf, np.inf, 5, np.inf, np.inf, np.inf])
        primitive_func = AbsoluteDiff().get_function()
        given_answer = primitive_func(data)
        np.testing.assert_array_equal(given_answer, answer)

    def test_raises(self):
        with pytest.raises(ValueError):
            AbsoluteDiff(method="invalid")



from datetime import datetime

import holidays
import numpy as np
import pandas as pd
import pytest
from packaging.version import parse

from featuretools.primitives import DistanceToHoliday


def test_distanceholiday():
    distance_to_holiday = DistanceToHoliday("New Year's Day")
    dates = pd.Series(
        [
            datetime(2010, 1, 1),
            datetime(2012, 5, 31),
            datetime(2017, 7, 31),
            datetime(2020, 12, 31),
        ],
    )

    expected = [0, -151, 154, 1]
    output = distance_to_holiday(dates).tolist()
    np.testing.assert_array_equal(output, expected)


def test_holiday_out_of_range():
    date_to_holiday = DistanceToHoliday("Boxing Day", country="Canada")

    array = pd.Series(
        [
            datetime(2010, 1, 1),
            datetime(2012, 5, 31),
            datetime(2017, 7, 31),
            datetime(2020, 12, 31),
        ],
    )
    days_to_boxing_day = -157 if parse(holidays.__version__) >= parse("0.15.0") else 209
    edge_case_first_day_of_year = (
        -6 if parse(holidays.__version__) >= parse("0.17.0") else np.nan
    )
    edge_case_last_day_of_year = (
        -5 if parse(holidays.__version__) >= parse("0.17.0") else np.nan
    )
    answer = pd.Series(
        [
            edge_case_first_day_of_year,
            days_to_boxing_day,
            148,
            edge_case_last_day_of_year,
        ],
    )
    pd.testing.assert_series_equal(date_to_holiday(array), answer, check_names=False)


def test_unknown_country_error():
    error_text = r"must be one of the available countries.*"
    with pytest.raises(ValueError, match=error_text):
        DistanceToHoliday("Victoria Day", country="UNK")


def test_unknown_holiday_error():
    error_text = r"must be one of the available holidays.*"
    with pytest.raises(ValueError, match=error_text):
        DistanceToHoliday("Alteryx Day")


def test_nat():
    date_to_holiday = DistanceToHoliday("New Year's Day")
    case = pd.Series(
        [
            "2010-01-01",
            "NaT",
            "2012-05-31",
            "NaT",
        ],
    ).astype("datetime64[ns]")
    answer = [0, np.nan, -151, np.nan]
    given_answer = date_to_holiday(case).astype("float")
    np.testing.assert_array_equal(given_answer, answer)


def test_valid_country():
    distance_to_holiday = DistanceToHoliday("Victoria Day", country="Canada")
    case = pd.Series(
        [
            "2010-01-01",
            "2012-05-31",
            "2017-07-31",
            "2020-12-31",
        ],
    ).astype("datetime64[ns]")
    answer = [143, -10, -70, 144]
    given_answer = distance_to_holiday(case).astype("float")
    np.testing.assert_array_equal(given_answer, answer)


def test_with_timezone_aware_datetimes():
    df = pd.DataFrame(
        {
            "non_timezone_aware_with_time": pd.date_range(
                "2018-07-03 09:00",
                periods=3,
            ),
            "non_timezone_aware_no_time": pd.date_range("2018-07-03", periods=3),
            "timezone_aware_with_time": pd.date_range(
                "2018-07-03 09:00",
                periods=3,
            ).tz_localize(tz="US/Eastern"),
            "timezone_aware_no_time": pd.date_range(
                "2018-07-03",
                periods=3,
            ).tz_localize(tz="US/Eastern"),
        },
    )

    distance_to_holiday = DistanceToHoliday("Independence Day", country="US")
    expected = [1, 0, -1]
    for col in df.columns:
        actual = distance_to_holiday(df[col])
        np.testing.assert_array_equal(actual, expected)



import pandas as pd

from featuretools import (
    AggregationFeature,
    Feature,
    IdentityFeature,
    TransformFeature,
    __version__,
)
from featuretools.entityset.deserialize import description_to_entityset
from featuretools.feature_base.features_serializer import FeaturesSerializer
from featuretools.primitives import (
    Count,
    Max,
    MultiplyNumericScalar,
    NMostCommon,
    NumUnique,
)
from featuretools.primitives.utils import serialize_primitive
from featuretools.version import FEATURES_SCHEMA_VERSION


def test_single_feature(es):
    feature = IdentityFeature(es["log"].ww["value"])
    serializer = FeaturesSerializer([feature])

    expected = {
        "ft_version": __version__,
        "schema_version": FEATURES_SCHEMA_VERSION,
        "entityset": es.to_dictionary(),
        "feature_list": [feature.unique_name()],
        "feature_definitions": {feature.unique_name(): feature.to_dictionary()},
        "primitive_definitions": {},
    }

    _compare_feature_dicts(expected, serializer.to_dict())


def test_base_features_in_list(es):
    value = IdentityFeature(es["log"].ww["value"])
    max_feature = AggregationFeature(value, "sessions", Max)
    features = [max_feature, value]
    serializer = FeaturesSerializer(features)

    expected = {
        "ft_version": __version__,
        "schema_version": FEATURES_SCHEMA_VERSION,
        "entityset": es.to_dictionary(),
        "feature_list": [max_feature.unique_name(), value.unique_name()],
        "feature_definitions": {
            max_feature.unique_name(): max_feature.to_dictionary(),
            value.unique_name(): value.to_dictionary(),
        },
    }
    expected["primitive_definitions"] = {
        "0": serialize_primitive(max_feature.primitive),
    }
    expected["feature_definitions"][max_feature.unique_name()]["arguments"][
        "primitive"
    ] = "0"

    actual = serializer.to_dict()
    _compare_feature_dicts(expected, actual)


def test_multi_output_features(es):
    product_id = IdentityFeature(es["log"].ww["product_id"])
    threecommon = NMostCommon()
    num_unique = NumUnique()
    tc = Feature(product_id, parent_dataframe_name="sessions", primitive=threecommon)

    features = [tc, product_id]
    for i in range(3):
        features.append(
            Feature(
                tc[i],
                parent_dataframe_name="customers",
                primitive=num_unique,
            ),
        )
        features.append(tc[i])

    serializer = FeaturesSerializer(features)

    flist = [feat.unique_name() for feat in features]
    fd = [feat.to_dictionary() for feat in features]
    fdict = dict(zip(flist, fd))

    expected = {
        "ft_version": __version__,
        "schema_version": FEATURES_SCHEMA_VERSION,
        "entityset": es.to_dictionary(),
        "feature_list": flist,
        "feature_definitions": fdict,
    }
    expected["primitive_definitions"] = {
        "0": serialize_primitive(tc.primitive),
        "1": serialize_primitive(features[2].primitive),
    }

    expected["feature_definitions"][flist[0]]["arguments"]["primitive"] = "0"
    expected["feature_definitions"][flist[2]]["arguments"]["primitive"] = "1"
    expected["feature_definitions"][flist[4]]["arguments"]["primitive"] = "1"
    expected["feature_definitions"][flist[6]]["arguments"]["primitive"] = "1"

    actual = serializer.to_dict()
    _compare_feature_dicts(expected, actual)


def test_base_features_not_in_list(es):
    max_primitive = Max()
    mult_primitive = MultiplyNumericScalar(value=2)
    value = IdentityFeature(es["log"].ww["value"])
    value_x2 = TransformFeature(value, mult_primitive)
    max_feature = AggregationFeature(value_x2, "sessions", max_primitive)
    features = [max_feature]
    serializer = FeaturesSerializer(features)

    expected = {
        "ft_version": __version__,
        "schema_version": FEATURES_SCHEMA_VERSION,
        "entityset": es.to_dictionary(),
        "feature_list": [max_feature.unique_name()],
        "feature_definitions": {
            max_feature.unique_name(): max_feature.to_dictionary(),
            value_x2.unique_name(): value_x2.to_dictionary(),
            value.unique_name(): value.to_dictionary(),
        },
    }
    expected["primitive_definitions"] = {
        "0": serialize_primitive(max_feature.primitive),
        "1": serialize_primitive(value_x2.primitive),
    }
    expected["feature_definitions"][max_feature.unique_name()]["arguments"][
        "primitive"
    ] = "0"
    expected["feature_definitions"][value_x2.unique_name()]["arguments"][
        "primitive"
    ] = "1"

    actual = serializer.to_dict()
    _compare_feature_dicts(expected, actual)


def test_where_feature_dependency(es):
    max_primitive = Max()
    value = IdentityFeature(es["log"].ww["value"])
    is_purchased = IdentityFeature(es["log"].ww["purchased"])
    max_feature = AggregationFeature(
        value,
        "sessions",
        max_primitive,
        where=is_purchased,
    )
    features = [max_feature]
    serializer = FeaturesSerializer(features)

    expected = {
        "ft_version": __version__,
        "schema_version": FEATURES_SCHEMA_VERSION,
        "entityset": es.to_dictionary(),
        "feature_list": [max_feature.unique_name()],
        "feature_definitions": {
            max_feature.unique_name(): max_feature.to_dictionary(),
            value.unique_name(): value.to_dictionary(),
            is_purchased.unique_name(): is_purchased.to_dictionary(),
        },
    }
    expected["primitive_definitions"] = {
        "0": serialize_primitive(max_feature.primitive),
    }
    expected["feature_definitions"][max_feature.unique_name()]["arguments"][
        "primitive"
    ] = "0"

    actual = serializer.to_dict()
    _compare_feature_dicts(expected, actual)


def test_feature_use_previous_pd_timedelta(es):
    value = IdentityFeature(es["log"].ww["id"])
    td = pd.Timedelta(12, "W")
    count_primitive = Count()
    count_feature = AggregationFeature(
        value,
        "customers",
        count_primitive,
        use_previous=td,
    )
    features = [count_feature, value]
    serializer = FeaturesSerializer(features)

    expected = {
        "ft_version": __version__,
        "schema_version": FEATURES_SCHEMA_VERSION,
        "entityset": es.to_dictionary(),
        "feature_list": [count_feature.unique_name(), value.unique_name()],
        "feature_definitions": {
            count_feature.unique_name(): count_feature.to_dictionary(),
            value.unique_name(): value.to_dictionary(),
        },
    }
    expected["primitive_definitions"] = {
        "0": serialize_primitive(count_feature.primitive),
    }
    expected["feature_definitions"][count_feature.unique_name()]["arguments"][
        "primitive"
    ] = "0"

    actual = serializer.to_dict()
    _compare_feature_dicts(expected, actual)


def test_feature_use_previous_pd_dateoffset(es):
    value = IdentityFeature(es["log"].ww["id"])
    do = pd.DateOffset(months=3)
    count_primitive = Count()
    count_feature = AggregationFeature(
        value,
        "customers",
        count_primitive,
        use_previous=do,
    )
    features = [count_feature, value]
    serializer = FeaturesSerializer(features)

    expected = {
        "ft_version": __version__,
        "schema_version": FEATURES_SCHEMA_VERSION,
        "entityset": es.to_dictionary(),
        "feature_list": [count_feature.unique_name(), value.unique_name()],
        "feature_definitions": {
            count_feature.unique_name(): count_feature.to_dictionary(),
            value.unique_name(): value.to_dictionary(),
        },
    }
    expected["primitive_definitions"] = {
        "0": serialize_primitive(count_feature.primitive),
    }
    expected["feature_definitions"][count_feature.unique_name()]["arguments"][
        "primitive"
    ] = "0"

    actual = serializer.to_dict()
    _compare_feature_dicts(expected, actual)

    value = IdentityFeature(es["log"].ww["id"])
    do = pd.DateOffset(months=3, days=2, minutes=30)
    count_feature = AggregationFeature(
        value,
        "customers",
        count_primitive,
        use_previous=do,
    )
    features = [count_feature, value]
    serializer = FeaturesSerializer(features)

    expected = {
        "ft_version": __version__,
        "schema_version": FEATURES_SCHEMA_VERSION,
        "entityset": es.to_dictionary(),
        "feature_list": [count_feature.unique_name(), value.unique_name()],
        "feature_definitions": {
            count_feature.unique_name(): count_feature.to_dictionary(),
            value.unique_name(): value.to_dictionary(),
        },
    }
    expected["primitive_definitions"] = {
        "0": serialize_primitive(count_feature.primitive),
    }
    expected["feature_definitions"][count_feature.unique_name()]["arguments"][
        "primitive"
    ] = "0"
    actual = serializer.to_dict()
    _compare_feature_dicts(expected, actual)


def _compare_feature_dicts(a_dict, b_dict):
    # We can't compare entityset dictionaries because column lists are not
    # guaranteed to be in the same order.
    es_a = description_to_entityset(a_dict.pop("entityset"))
    es_b = description_to_entityset(b_dict.pop("entityset"))
    assert es_a == es_b

    assert a_dict == b_dict



from featuretools.tests.entry_point_tests.utils import (
    _import_featuretools,
    _install_featuretools_plugin,
    _uninstall_featuretools_plugin,
)


def test_plugin_warning():
    _install_featuretools_plugin()
    warning = _import_featuretools("warning").stdout.decode()
    debug = _import_featuretools("debug").stdout.decode()
    _uninstall_featuretools_plugin()

    message = (
        "Featuretools failed to load plugin module from library featuretools_plugin"
    )
    traceback = "NotImplementedError: plugin not implemented"

    assert message in warning
    assert traceback not in warning
    assert message in debug
    assert traceback in debug



import numpy as np
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Boolean, BooleanNullable

from featuretools.primitives.base.transform_primitive_base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class Or(TransformPrimitive):
    """Performs element-wise logical OR of two lists.

    Description:
        Given a list of booleans X and a list of booleans Y,
        determine whether each value in X is `True`, or
        whether its corresponding value in Y is `True`.

    Examples:
        >>> _or = Or()
        >>> _or([False, True, False], [True, True, False]).tolist()
        [True, True, False]
    """

    name = "or"
    input_types = [
        [
            ColumnSchema(logical_type=BooleanNullable),
            ColumnSchema(logical_type=BooleanNullable),
        ],
        [ColumnSchema(logical_type=Boolean), ColumnSchema(logical_type=Boolean)],
        [
            ColumnSchema(logical_type=Boolean),
            ColumnSchema(logical_type=BooleanNullable),
        ],
        [
            ColumnSchema(logical_type=BooleanNullable),
            ColumnSchema(logical_type=Boolean),
        ],
    ]
    return_type = ColumnSchema(logical_type=BooleanNullable)
    commutative = True
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]
    description_template = "whether {} is true or {} is true"

    def get_function(self):
        return np.logical_or

    def generate_name(self, base_feature_names):
        return "OR(%s, %s)" % (base_feature_names[0], base_feature_names[1])



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Datetime, Ordinal

from featuretools.primitives.base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class Week(TransformPrimitive):
    """Determines the week of the year from a datetime.

    Description:
        Returns the week of the year from a datetime value. The first week
        of the year starts on January 1, and week numbers increment each
        Monday.

    Examples:
        >>> from datetime import datetime
        >>> dates = [datetime(2019, 1, 3),
        ...          datetime(2019, 6, 17, 11, 10, 50),
        ...          datetime(2019, 11, 30, 19, 45, 15)]
        >>> week = Week()
        >>> week(dates).tolist()
        [1, 25, 48]
    """

    name = "week"
    input_types = [ColumnSchema(logical_type=Datetime)]
    return_type = ColumnSchema(
        logical_type=Ordinal(order=list(range(1, 54))),
        semantic_tags={"category"},
    )
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]
    description_template = "the week of the year of {}"

    def get_function(self):
        def week(vals):
            if hasattr(vals.dt, "isocalendar"):
                return vals.dt.isocalendar().week
            else:
                return vals.dt.week

        return week



import re

import numpy as np
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import IntegerNullable, NaturalLanguage

from featuretools.primitives.base import TransformPrimitive


class CountString(TransformPrimitive):
    """Determines how many times a given string shows up in a text field.

    Args:
        string (str): The string to determine the count of. Defaults to
            the word "the".
        ignore_case (bool): Determines if case of the string should be
            considered or not. Defaults to true.
        ignore_non_alphanumeric (bool): Determines if non-alphanumeric
            characters should be used in the search. Defaults to False.
        is_regex (bool): Defines if the string argument is a regex or not.
            Defaults to False.
        match_whole_words_only (bool): Determines if whole words should be
            matched or not. For example searching for word `the` against
            `then, the, there` should only return `the` if this argument
            was True. Defaults to False.
    Examples:
        >>> count_string = CountString(string="the")
        >>> count_string(["The problem was difficult.",
        ...               "He was there.",
        ...               "The girl went to the store."]).tolist()
        [1.0, 1.0, 2.0]
        >>> # Match case of string
        >>> count_string_ignore_case = CountString(string="the", ignore_case=False)
        >>> count_string_ignore_case(["The problem was difficult.",
        ...                           "He was there.",
        ...                           "The girl went to the store."]).tolist()
        [0.0, 1.0, 1.0]
        >>> # Ignore non-alphanumeric characters in the search
        >>> count_string_ignore_non_alphanumeric = CountString(string="the",
        ...                                                    ignore_non_alphanumeric=True)
        >>> count_string_ignore_non_alphanumeric(["Th*/e problem was difficult.",
        ...                                       "He was there.",
        ...                                       "The girl went to the store."]).tolist()
        [1.0, 1.0, 2.0]
        >>> # Specify the string as a regex
        >>> count_string_is_regex = CountString(string="t.e", is_regex=True)
        >>> count_string_is_regex(["The problem was difficult.",
        ...                        "He was there.",
        ...                        "The girl went to the store."]).tolist()
        [1.0, 1.0, 2.0]
        >>> # Match whole words only
        >>> count_string_match_whole_words_only = CountString(string="the",
        ...                                                   match_whole_words_only=True)
        >>> count_string_match_whole_words_only(["The problem was difficult.",
        ...                                      "He was there.",
        ...                                      "The girl went to the store."]).tolist()
        [1.0, 0.0, 2.0]
    """

    name = "count_string"
    input_types = [ColumnSchema(logical_type=NaturalLanguage)]
    return_type = ColumnSchema(logical_type=IntegerNullable, semantic_tags={"numeric"})

    def __init__(
        self,
        string="the",
        ignore_case=True,
        ignore_non_alphanumeric=False,
        is_regex=False,
        match_whole_words_only=False,
    ):
        self.string = string
        self.ignore_case = ignore_case
        self.ignore_non_alphanumeric = ignore_non_alphanumeric
        self.match_whole_words_only = match_whole_words_only
        self.is_regex = is_regex

        # we don't want to strip non alphanumeric characters from the pattern
        # ie h.ll. should match "hello" so we can't strip the dots to make hll
        if not is_regex:
            self.pattern = re.escape(self.process_text(string))
        else:
            self.pattern = string
            if ignore_case:
                self.pattern = self.pattern.lower()

        # \b\b.*\b\b is the same as \b.*\b so we don't have to check if
        # the pattern is given to us as regex and if it already has leading
        # and trailing \b's
        if match_whole_words_only:
            self.pattern = "\\b" + self.pattern + "\\b"

    def process_text(self, text):
        if self.ignore_non_alphanumeric:
            text = re.sub("[^0-9a-zA-Z ]+", "", text)
        if self.ignore_case:
            text = text.lower()
        return text

    def get_function(self):
        def count_string(words):
            if not isinstance(words, str):
                return np.nan
            words = self.process_text(words)
            return len(re.findall(self.pattern, words))

        return np.vectorize(count_string, otypes=[float])



from featuretools.primitives.standard.transform.email.email_address_to_domain import (
    EmailAddressToDomain,
)
from featuretools.primitives.standard.transform.email.is_free_email_domain import (
    IsFreeEmailDomain,
)



import numpy as np
import pandas as pd
import pytest

from featuretools.primitives import SameAsPrevious


class TestSameAsPrevious:
    def test_ints(self):
        primitive_func = SameAsPrevious().get_function()
        array = pd.Series([1, 2, 2, 3, 2], dtype="int64")
        answer = primitive_func(array)
        correct_answer = pd.Series([False, False, True, False, False])
        pd.testing.assert_series_equal(answer, correct_answer)

    def test_int64(self):
        primitive_func = SameAsPrevious().get_function()
        array = pd.Series([1, 2, 2, 3, 2], dtype="Int64")
        answer = primitive_func(array)
        correct_answer = pd.Series([False, False, True, False, False], dtype="boolean")
        pd.testing.assert_series_equal(answer, correct_answer)

    def test_floats(self):
        primitive_func = SameAsPrevious().get_function()
        array = pd.Series([1.0, 2.5, 2.5, 3.0, 2.0], dtype="float64")
        answer = primitive_func(array)
        correct_answer = pd.Series([False, False, True, False, False])
        pd.testing.assert_series_equal(answer, correct_answer)

    def test_mixed(self):
        primitive_func = SameAsPrevious().get_function()
        array = pd.Series([1, 2, 2.0, 3, 2.0], dtype="float64")
        answer = primitive_func(array)
        correct_answer = pd.Series([False, False, True, False, False])
        np.testing.assert_array_equal(answer, correct_answer)

    def test_nan(self):
        primitive_instance = SameAsPrevious()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([1, np.nan, 3, np.nan, 2], dtype="float64")
        answer = primitive_func(array)
        correct_answer = pd.Series([False, True, False, True, False])
        np.testing.assert_array_equal(answer, correct_answer)

    def test_all_nan(self):
        primitive_instance = SameAsPrevious()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([np.nan, np.nan, np.nan, np.nan], dtype="float64")
        answer = primitive_func(array)
        correct_answer = pd.Series([False, False, False, False])
        np.testing.assert_array_equal(answer, correct_answer)

    def test_inf(self):
        primitive_instance = SameAsPrevious()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([1, np.inf, 3, np.inf, 2], dtype="float64")
        answer = primitive_func(array)
        correct_answer = pd.Series([False, False, False, False, False])
        np.testing.assert_array_equal(answer, correct_answer)

    def test_all_inf(self):
        primitive_instance = SameAsPrevious()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([np.inf, np.inf, np.inf, np.inf], dtype="float64")
        answer = primitive_func(array)
        correct_answer = pd.Series([False, True, True, True])
        np.testing.assert_array_equal(answer, correct_answer)

    def test_fill_method_bfill(self):
        primitive_instance = SameAsPrevious(fill_method="bfill")
        primitive_func = primitive_instance.get_function()
        array = pd.Series([1, np.nan, 3, 2, 2], dtype="float64")
        answer = primitive_func(array)
        correct_answer = pd.Series([False, False, True, False, True])
        np.testing.assert_array_equal(answer, correct_answer)

    def test_fill_method_bfill_with_limit(self):
        primitive_instance = SameAsPrevious(fill_method="bfill", limit=2)
        primitive_func = primitive_instance.get_function()
        array = pd.Series([1, np.nan, np.nan, np.nan, 2, 3], dtype="float64")
        answer = primitive_func(array)
        correct_answer = pd.Series([False, False, False, True, True, False])
        np.testing.assert_array_equal(answer, correct_answer)

    def test_raises(self):
        with pytest.raises(ValueError):
            SameAsPrevious(fill_method="invalid")



import warnings

import pandas as pd
from woodwork.column_schema import ColumnSchema

from featuretools.primitives.base import TransformPrimitive


class NumericLag(TransformPrimitive):
    """Shifts an array of values by a specified number of periods.

    Args:
        periods (int): The number of periods by which to shift the input.
            Default is 1. Periods correspond to rows.

        fill_value (int, float, optional): The value to use to fill in
            the gaps left after shifting the input. Default is None.

    Examples:
        >>> lag = NumericLag()
        >>> lag(pd.Series(pd.date_range(start="2020-01-01", periods=5, freq='D')), [1, 2, 3, 4, 5]).tolist()
        [nan, 1.0, 2.0, 3.0, 4.0]

        You can specify the number of periods to shift the values

        >>> lag_periods = NumericLag(periods=3)
        >>> lag_periods(pd.Series(pd.date_range(start="2020-01-01", periods=5, freq='D')), [1, 2, 3, 4, 5]).tolist()
        [nan, nan, nan, 1.0, 2.0]

        You can specify the fill value to use

        >>> lag_fill_value = NumericLag(fill_value=100)
        >>> lag_fill_value(pd.Series(pd.date_range(start="2020-01-01", periods=4, freq='D')), [1, 2, 3, 4]).tolist()
        [100, 1, 2, 3]
    """

    name = "numeric_lag"
    input_types = [
        ColumnSchema(semantic_tags={"time_index"}),
        ColumnSchema(semantic_tags={"numeric"}),
    ]
    return_type = ColumnSchema(semantic_tags={"numeric"})
    uses_full_dataframe = True

    def __init__(self, periods=1, fill_value=None):
        self.periods = periods
        self.fill_value = fill_value
        warnings.warn(
            "NumericLag is deprecated and will be removed in a future version. Please use the 'Lag' primitive instead.",
            FutureWarning,
        )

    def get_function(self):
        def lag(time_index, numeric):
            x = pd.Series(numeric.values, index=time_index.values)
            return x.shift(periods=self.periods, fill_value=self.fill_value).values

        return lag



from woodwork.column_schema import ColumnSchema

from featuretools.primitives.base.transform_primitive_base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class ModuloByFeature(TransformPrimitive):
    """Computes the modulo of a scalar by each element in a list.

    Description:
        Given a list of numeric values and a scalar, return the
        modulo, or remainder of the scalar after being divided
        by each value.

    Examples:
        >>> modulo_by_feature = ModuloByFeature(value=2)
        >>> modulo_by_feature([4, 1, 2]).tolist()
        [2, 0, 0]
    """

    name = "modulo_by_feature"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(semantic_tags={"numeric"})
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]

    def __init__(self, value=1):
        self.value = value
        self.description_template = "the remainder after dividing {} by {{}}".format(
            self.value,
        )

    def get_function(self):
        def modulo_by_feature(vals):
            return self.value % vals

        return modulo_by_feature

    def generate_name(self, base_feature_names):
        return "%s %% %s" % (str(self.value), base_feature_names[0])



import numpy as np
import pandas as pd
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Categorical

from featuretools.primitives.base import AggregationPrimitive


class NMostCommonFrequency(AggregationPrimitive):
    """Determines the frequency of the n most common items.

    Args:
        n (int): defines "n" in "n most common". Defaults to
            3.
        skipna (bool): Determines if to use NA/null values.
            Defaults to True to skip NA/null.

    Description:
        Given a list, find the n most common items, and return a series
        showing the frequency of each item. If the list has less than n unique
        values, the resulting series will be padded with nan.

    Examples:
        >>> n_most_common_frequency = NMostCommonFrequency()
        >>> n_most_common_frequency([1, 1, 1, 2, 2, 3, 4, 4]).to_list()
        [3, 2, 2]

        We can increase n to include more items.

        >>> n_most_common_frequency = NMostCommonFrequency(4)
        >>> n_most_common_frequency([1, 1, 1, 2, 2, 3, 4, 4]).to_list()
        [3, 2, 2, 1]

        `NaN`s are skipped by default.

        >>> n_most_common_frequency = NMostCommonFrequency(3)
        >>> n_most_common_frequency([1, 1, 1, 2, 2, 3, 4, 4, None, None, None]).to_list()
        [3, 2, 2]

        However, the way `NaN`s are treated can be controlled.

        >>> n_most_common_frequency = NMostCommonFrequency(3, skipna=False)
        >>> n_most_common_frequency([1, 1, 1, 2, 2, 3, 4, 4, None, None, None]).to_list()
        [3, 3, 2]
    """

    name = "n_most_common_frequency"
    input_types = [ColumnSchema(semantic_tags={"category"})]
    return_type = ColumnSchema(logical_type=Categorical, semantic_tags={"category"})

    def __init__(self, n=3, skipna=True):
        self.n = n
        self.number_output_features = n
        self.skipna = skipna

    def get_function(self):
        def n_most_common_frequency(data, n=self.n):
            frequencies = data.value_counts(dropna=self.skipna)
            n_most_common = frequencies.iloc[0:n]
            nan_add = n - frequencies.shape[0]
            if nan_add > 0:
                n_most_common = pd.concat(
                    [n_most_common, pd.Series([np.nan] * nan_add)],
                )
            return n_most_common

        return n_most_common_frequency



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Datetime, Timedelta

from featuretools.primitives.standard.transform.numeric.diff import Diff


class DiffDatetime(Diff):
    """Computes the timedelta between a datetime in a list and the
    previous datetime in that list.

    Args:
        periods (int): The number of periods by which to shift the index row.
            Default is 0. Periods correspond to rows.

    Description:
        Given a list of datetimes, compute the difference from the previous
        item in the list. The result for the first element of the list will
        always be `NaT`.

    Examples:
        >>> from datetime import datetime
        >>> dt_values = [datetime(2019, 3, 1), datetime(2019, 6, 30), datetime(2019, 11, 17), datetime(2020, 1, 30), datetime(2020, 3, 11)]
        >>> diff_dt = DiffDatetime()
        >>> diff_dt(dt_values).tolist()
        [NaT, Timedelta('121 days 00:00:00'), Timedelta('140 days 00:00:00'), Timedelta('74 days 00:00:00'), Timedelta('41 days 00:00:00')]

        You can specify the number of periods to shift the values

        >>> diff_dt_periods = DiffDatetime(periods = 1)
        >>> diff_dt_periods(dt_values).tolist()
        [NaT, NaT, Timedelta('121 days 00:00:00'), Timedelta('140 days 00:00:00'), Timedelta('74 days 00:00:00')]
    """

    name = "diff_datetime"
    input_types = [ColumnSchema(logical_type=Datetime)]
    return_type = ColumnSchema(logical_type=Timedelta)
    uses_full_dataframe = True
    description_template = "the difference from the previous value of {}"

    def __init__(self, periods=0):
        super().__init__(periods)



def is_valid_input(candidate, template):
    """Checks if a candidate schema should be considered a match for a template schema"""
    if template.logical_type is not None and not isinstance(
        candidate.logical_type,
        type(template.logical_type),
    ):
        return False
    if len(template.semantic_tags - candidate.semantic_tags):
        return False
    return True



import pandas as pd
import pytest

from featuretools import EntitySet, calculate_feature_matrix, dfs
from featuretools.feature_base import Feature, IdentityFeature
from featuretools.primitives import NMostCommon
from featuretools.synthesis import encode_features


def test_encodes_features(pd_es):
    f1 = IdentityFeature(pd_es["log"].ww["product_id"])
    f2 = IdentityFeature(pd_es["log"].ww["purchased"])
    f3 = IdentityFeature(pd_es["log"].ww["value"])

    features = [f1, f2, f3]
    feature_matrix = calculate_feature_matrix(
        features,
        pd_es,
        instance_ids=[0, 1, 2, 3, 4, 5],
    )

    _, features_encoded = encode_features(feature_matrix, features)
    assert len(features_encoded) == 6

    _, features_encoded = encode_features(feature_matrix, features, top_n=2)
    assert len(features_encoded) == 5

    _, features_encoded = encode_features(
        feature_matrix,
        features,
        include_unknown=False,
    )
    assert len(features_encoded) == 5


def test_dask_errors_encode_features(dask_es):
    f1 = IdentityFeature(dask_es["log"].ww["product_id"])
    f2 = IdentityFeature(dask_es["log"].ww["purchased"])
    f3 = IdentityFeature(dask_es["log"].ww["value"])

    features = [f1, f2, f3]
    feature_matrix = calculate_feature_matrix(
        features,
        dask_es,
        instance_ids=[0, 1, 2, 3, 4, 5],
    )
    error_text = "feature_matrix must be a Pandas DataFrame"

    with pytest.raises(TypeError, match=error_text):
        encode_features(feature_matrix, features)


def test_inplace_encodes_features(pd_es):
    f1 = IdentityFeature(pd_es["log"].ww["product_id"])

    features = [f1]
    feature_matrix = calculate_feature_matrix(
        features,
        pd_es,
        instance_ids=[0, 1, 2, 3, 4, 5],
    )

    feature_matrix_shape = feature_matrix.shape
    feature_matrix_encoded, _ = encode_features(feature_matrix, features)
    assert feature_matrix_encoded.shape != feature_matrix_shape
    assert feature_matrix.shape == feature_matrix_shape

    # inplace they should be the same
    feature_matrix_encoded, _ = encode_features(feature_matrix, features, inplace=True)
    assert feature_matrix_encoded.shape == feature_matrix.shape


def test_to_encode_features(pd_es):
    f1 = IdentityFeature(pd_es["log"].ww["product_id"])
    f2 = IdentityFeature(pd_es["log"].ww["value"])
    f3 = IdentityFeature(pd_es["log"].ww["datetime"])

    features = [f1, f2, f3]
    feature_matrix = calculate_feature_matrix(
        features,
        pd_es,
        instance_ids=[0, 1, 2, 3, 4, 5],
    )

    feature_matrix_encoded, _ = encode_features(feature_matrix, features)
    feature_matrix_encoded_shape = feature_matrix_encoded.shape

    # to_encode should keep product_id as a string and datetime as a date,
    # and not have the same shape as previous encoded matrix due to fewer encoded features
    to_encode = []
    feature_matrix_encoded, _ = encode_features(
        feature_matrix,
        features,
        to_encode=to_encode,
    )
    assert feature_matrix_encoded_shape != feature_matrix_encoded.shape
    assert feature_matrix_encoded["datetime"].dtype == "datetime64[ns]"
    assert feature_matrix_encoded["product_id"].dtype == "category"

    to_encode = ["value"]
    feature_matrix_encoded, _ = encode_features(
        feature_matrix,
        features,
        to_encode=to_encode,
    )
    assert feature_matrix_encoded_shape != feature_matrix_encoded.shape
    assert feature_matrix_encoded["datetime"].dtype == "datetime64[ns]"
    assert feature_matrix_encoded["product_id"].dtype == "category"


def test_encode_features_handles_pass_columns(pd_es):
    f1 = IdentityFeature(pd_es["log"].ww["product_id"])
    f2 = IdentityFeature(pd_es["log"].ww["value"])

    features = [f1, f2]
    cutoff_time = pd.DataFrame(
        {
            "instance_id": range(6),
            "time": pd_es["log"]["datetime"][0:6],
            "label": [i % 2 for i in range(6)],
        },
        columns=["instance_id", "time", "label"],
    )
    feature_matrix = calculate_feature_matrix(features, pd_es, cutoff_time)

    assert "label" in feature_matrix.columns

    feature_matrix_encoded, _ = encode_features(feature_matrix, features)
    feature_matrix_encoded_shape = feature_matrix_encoded.shape

    # to_encode should keep product_id as a string, and not create 3 additional columns
    to_encode = []
    feature_matrix_encoded, _ = encode_features(
        feature_matrix,
        features,
        to_encode=to_encode,
    )
    assert feature_matrix_encoded_shape != feature_matrix_encoded.shape

    to_encode = ["value"]
    feature_matrix_encoded, _ = encode_features(
        feature_matrix,
        features,
        to_encode=to_encode,
    )
    assert feature_matrix_encoded_shape != feature_matrix_encoded.shape

    assert "label" in feature_matrix_encoded.columns


def test_encode_features_catches_features_mismatch(pd_es):
    f1 = IdentityFeature(pd_es["log"].ww["product_id"])
    f2 = IdentityFeature(pd_es["log"].ww["value"])
    f3 = IdentityFeature(pd_es["log"].ww["session_id"])

    features = [f1, f2]
    cutoff_time = pd.DataFrame(
        {
            "instance_id": range(6),
            "time": pd_es["log"]["datetime"][0:6],
            "label": [i % 2 for i in range(6)],
        },
        columns=["instance_id", "time", "label"],
    )
    feature_matrix = calculate_feature_matrix(features, pd_es, cutoff_time)

    assert "label" in feature_matrix.columns

    error_text = "Feature session_id not found in feature matrix"
    with pytest.raises(AssertionError, match=error_text):
        encode_features(feature_matrix, [f1, f3])


def test_encode_unknown_features():
    # Dataframe with categorical column with "unknown" string
    df = pd.DataFrame({"category": ["unknown", "b", "c", "d", "e"]}).astype(
        {"category": "category"},
    )

    pd_es = EntitySet("test")
    pd_es.add_dataframe(
        dataframe_name="a",
        dataframe=df,
        index="index",
        make_index=True,
    )
    features, feature_defs = dfs(
        entityset=pd_es,
        target_dataframe_name="a",
        max_depth=1,
    )

    # Specify unknown token for replacement
    features_enc, _ = encode_features(features, feature_defs, include_unknown=True)
    assert list(features_enc.columns) == [
        "category = unknown",
        "category = e",
        "category = d",
        "category = c",
        "category = b",
        "category is unknown",
    ]


def test_encode_features_topn(pd_es):
    topn = Feature(
        Feature(pd_es["log"].ww["product_id"]),
        parent_dataframe_name="customers",
        primitive=NMostCommon(n=3),
    )
    features, feature_defs = dfs(
        entityset=pd_es,
        instance_ids=[0, 1, 2],
        target_dataframe_name="customers",
        agg_primitives=[NMostCommon(n=3)],
    )
    features_enc, feature_defs_enc = encode_features(
        features,
        feature_defs,
        include_unknown=True,
    )
    assert topn.unique_name() in [feat.unique_name() for feat in feature_defs_enc]
    for name in topn.get_feature_names():
        assert name in features_enc.columns
        assert features_enc.columns.tolist().count(name) == 1


def test_encode_features_drop_first():
    df = pd.DataFrame({"category": ["ao", "b", "c", "d", "e"]}).astype(
        {"category": "category"},
    )
    pd_es = EntitySet("test")
    pd_es.add_dataframe(
        dataframe_name="a",
        dataframe=df,
        index="index",
        make_index=True,
    )
    features, feature_defs = dfs(
        entityset=pd_es,
        target_dataframe_name="a",
        max_depth=1,
    )
    features_enc, _ = encode_features(
        features,
        feature_defs,
        drop_first=True,
        include_unknown=False,
    )
    assert len(features_enc.columns) == 4

    features_enc, feature_defs = encode_features(
        features,
        feature_defs,
        top_n=3,
        drop_first=True,
        include_unknown=False,
    )

    assert len(features_enc.columns) == 2


def test_encode_features_handles_dictionary_input(pd_es):
    f1 = IdentityFeature(pd_es["log"].ww["product_id"])
    f2 = IdentityFeature(pd_es["log"].ww["purchased"])
    f3 = IdentityFeature(pd_es["log"].ww["session_id"])

    features = [f1, f2, f3]
    feature_matrix = calculate_feature_matrix(features, pd_es, instance_ids=range(16))
    feature_matrix_encoded, features_encoded = encode_features(feature_matrix, features)
    true_values = [
        "product_id = coke zero",
        "product_id = toothpaste",
        "product_id = car",
        "product_id = brown bag",
        "product_id = taco clock",
        "product_id = Haribo sugar-free gummy bears",
        "product_id is unknown",
        "purchased",
        "session_id = 0",
        "session_id = 1",
        "session_id = 4",
        "session_id = 3",
        "session_id = 5",
        "session_id = 2",
        "session_id is unknown",
    ]
    assert len(features_encoded) == 15
    for col in true_values:
        assert col in list(feature_matrix_encoded.columns)

    top_n_dict = {}
    feature_matrix_encoded, features_encoded = encode_features(
        feature_matrix,
        features,
        top_n=top_n_dict,
    )
    assert len(features_encoded) == 15
    for col in true_values:
        assert col in list(feature_matrix_encoded.columns)

    top_n_dict = {f1.get_name(): 4, f3.get_name(): 3}
    feature_matrix_encoded, features_encoded = encode_features(
        feature_matrix,
        features,
        top_n=top_n_dict,
    )
    assert len(features_encoded) == 10
    true_values = [
        "product_id = coke zero",
        "product_id = toothpaste",
        "product_id = car",
        "product_id = brown bag",
        "product_id is unknown",
        "purchased",
        "session_id = 0",
        "session_id = 1",
        "session_id = 4",
        "session_id is unknown",
    ]
    for col in true_values:
        assert col in list(feature_matrix_encoded.columns)

    feature_matrix_encoded, features_encoded = encode_features(
        feature_matrix,
        features,
        top_n=top_n_dict,
        include_unknown=False,
    )
    true_values = [
        "product_id = coke zero",
        "product_id = toothpaste",
        "product_id = car",
        "product_id = brown bag",
        "purchased",
        "session_id = 0",
        "session_id = 1",
        "session_id = 4",
    ]
    assert len(features_encoded) == 8
    for col in true_values:
        assert col in list(feature_matrix_encoded.columns)


def test_encode_features_matches_calculate_feature_matrix():
    df = pd.DataFrame({"category": ["b", "c", "d", "e"]}).astype(
        {"category": "category"},
    )

    pd_es = EntitySet("test")
    pd_es.add_dataframe(
        dataframe_name="a",
        dataframe=df,
        index="index",
        make_index=True,
    )
    features, feature_defs = dfs(
        entityset=pd_es,
        target_dataframe_name="a",
        max_depth=1,
    )

    features_enc, feature_defs_enc = encode_features(
        features,
        feature_defs,
        to_encode=["category"],
    )

    features_calc = calculate_feature_matrix(feature_defs_enc, entityset=pd_es)

    pd.testing.assert_frame_equal(features_enc, features_calc)
    assert features_calc.ww._schema == features_enc.ww._schema



from datetime import datetime

import pandas as pd
import pytest
from woodwork.logical_types import Categorical, Datetime, Integer

from featuretools.entityset.entityset import LTI_COLUMN_NAME
from featuretools.tests.testing_utils import to_pandas
from featuretools.utils.gen_utils import Library, import_or_none

dd = import_or_none("dask.dataframe")
ps = import_or_none("pyspark.pandas")


@pytest.fixture
def values_es(es):
    es.normalize_dataframe(
        "log",
        "values",
        "value",
        make_time_index=True,
        new_dataframe_time_index="value_time",
    )
    return es


@pytest.fixture
def true_values_lti():
    true_values_lti = pd.Series(
        [
            datetime(2011, 4, 10, 10, 41, 0),
            datetime(2011, 4, 9, 10, 31, 9),
            datetime(2011, 4, 9, 10, 31, 18),
            datetime(2011, 4, 9, 10, 31, 27),
            datetime(2011, 4, 10, 10, 40, 1),
            datetime(2011, 4, 10, 10, 41, 3),
            datetime(2011, 4, 9, 10, 30, 12),
            datetime(2011, 4, 10, 10, 41, 6),
            datetime(2011, 4, 9, 10, 30, 18),
            datetime(2011, 4, 9, 10, 30, 24),
            datetime(2011, 4, 10, 11, 10, 3),
        ],
    )
    return true_values_lti


@pytest.fixture
def true_sessions_lti():
    sessions_lti = pd.Series(
        [
            datetime(2011, 4, 9, 10, 30, 24),
            datetime(2011, 4, 9, 10, 31, 27),
            datetime(2011, 4, 9, 10, 40, 0),
            datetime(2011, 4, 10, 10, 40, 1),
            datetime(2011, 4, 10, 10, 41, 6),
            datetime(2011, 4, 10, 11, 10, 3),
        ],
    )
    return sessions_lti


@pytest.fixture
def wishlist_df():
    wishlist_df = pd.DataFrame(
        {
            "session_id": [0, 1, 2, 2, 3, 4, 5],
            "datetime": [
                datetime(2011, 4, 9, 10, 30, 15),
                datetime(2011, 4, 9, 10, 31, 30),
                datetime(2011, 4, 9, 10, 30, 30),
                datetime(2011, 4, 9, 10, 35, 30),
                datetime(2011, 4, 10, 10, 41, 0),
                datetime(2011, 4, 10, 10, 39, 59),
                datetime(2011, 4, 10, 11, 10, 2),
            ],
            "product_id": [
                "coke zero",
                "taco clock",
                "coke zero",
                "car",
                "toothpaste",
                "brown bag",
                "coke zero",
            ],
        },
    )
    return wishlist_df


@pytest.fixture
def extra_session_df(es):
    row_values = {"customer_id": 2, "device_name": "PC", "device_type": 0, "id": 6}
    row = pd.DataFrame(row_values, index=pd.Index([6], name="id"))
    df = to_pandas(es["sessions"])
    df = pd.concat([df, row]).sort_index()
    if es.dataframe_type == Library.DASK:
        df = dd.from_pandas(df, npartitions=3)
    elif es.dataframe_type == Library.SPARK:
        # Spark can't handle object dtypes
        df = df.astype("string")
        df = ps.from_pandas(df)
    return df


class TestLastTimeIndex(object):
    def test_leaf(self, es):
        es.add_last_time_indexes()
        log = es["log"]
        lti_name = log.ww.metadata.get("last_time_index")

        assert lti_name == LTI_COLUMN_NAME
        assert len(log[lti_name]) == 17

        log_df = to_pandas(log)

        for v1, v2 in zip(log_df[lti_name], log_df["datetime"]):
            assert (pd.isnull(v1) and pd.isnull(v2)) or v1 == v2

    def test_leaf_no_time_index(self, es):
        es.add_last_time_indexes()
        stores = es["stores"]
        true_lti = pd.Series([None for x in range(6)], dtype="datetime64[ns]")

        assert len(true_lti) == len(stores[LTI_COLUMN_NAME])

        stores_lti = to_pandas(stores[LTI_COLUMN_NAME])

        for v1, v2 in zip(stores_lti, true_lti):
            assert (pd.isnull(v1) and pd.isnull(v2)) or v1 == v2

    # TODO: possible issue with either normalize_dataframe or add_last_time_indexes
    def test_parent(self, values_es, true_values_lti):
        # test dataframe with time index and all instances in child dataframe
        if values_es.dataframe_type != Library.PANDAS:
            pytest.xfail(
                "possible issue with either normalize_dataframe or add_last_time_indexes",
            )
        values_es.add_last_time_indexes()
        values = values_es["values"]
        lti_name = values.ww.metadata.get("last_time_index")
        assert len(values[lti_name]) == 10
        sorted_lti = to_pandas(values[lti_name]).sort_index()
        for v1, v2 in zip(sorted_lti, true_values_lti):
            assert (pd.isnull(v1) and pd.isnull(v2)) or v1 == v2

    # TODO: fails with Dask, tests needs to be reworked
    def test_parent_some_missing(self, values_es, true_values_lti):
        # test dataframe with time index and not all instances have children
        if values_es.dataframe_type != Library.PANDAS:
            pytest.xfail("fails with Dask, tests needs to be reworked")
        values = values_es["values"]

        # add extra value instance with no children
        row_values = {
            "value": [21.0],
            "value_time": [pd.Timestamp("2011-04-10 11:10:02")],
        }
        # make sure index doesn't have same name as column to suppress pandas warning
        row = pd.DataFrame(row_values, index=pd.Index([21]))
        df = pd.concat([values, row])
        df = df.sort_values(by="value")
        df.index.name = None

        values_es.replace_dataframe(dataframe_name="values", df=df)
        values_es.add_last_time_indexes()
        # lti value should default to instance's time index
        true_values_lti[10] = pd.Timestamp("2011-04-10 11:10:02")
        true_values_lti[11] = pd.Timestamp("2011-04-10 11:10:03")

        values = values_es["values"]
        lti_name = values.ww.metadata.get("last_time_index")
        assert len(values[lti_name]) == 11
        sorted_lti = values[lti_name].sort_index()
        for v1, v2 in zip(sorted_lti, true_values_lti):
            assert (pd.isnull(v1) and pd.isnull(v2)) or v1 == v2

    def test_parent_no_time_index(self, es, true_sessions_lti):
        # test dataframe without time index and all instances have children
        es.add_last_time_indexes()
        sessions = es["sessions"]
        lti_name = sessions.ww.metadata.get("last_time_index")
        assert len(sessions[lti_name]) == 6
        sorted_lti = to_pandas(sessions[lti_name]).sort_index()
        for v1, v2 in zip(sorted_lti, true_sessions_lti):
            assert (pd.isnull(v1) and pd.isnull(v2)) or v1 == v2

    def test_parent_no_time_index_missing(
        self,
        es,
        extra_session_df,
        true_sessions_lti,
    ):
        # test dataframe without time index and not all instance have children

        # add session instance with no associated log instances
        es.replace_dataframe(dataframe_name="sessions", df=extra_session_df)
        es.add_last_time_indexes()
        # since sessions has no time index, default value is NaT
        true_sessions_lti[6] = pd.NaT
        sessions = es["sessions"]

        lti_name = sessions.ww.metadata.get("last_time_index")
        assert len(sessions[lti_name]) == 7
        sorted_lti = to_pandas(sessions[lti_name]).sort_index()
        for v1, v2 in zip(sorted_lti, true_sessions_lti):
            assert (pd.isnull(v1) and pd.isnull(v2)) or v1 == v2

    def test_multiple_children(self, es, wishlist_df, true_sessions_lti):
        if es.dataframe_type == Library.SPARK:
            pytest.xfail("Cannot make index on a Spark DataFrame")
        # test all instances in both children
        if es.dataframe_type == Library.DASK:
            wishlist_df = dd.from_pandas(wishlist_df, npartitions=2)
        logical_types = {
            "session_id": Integer,
            "datetime": Datetime,
            "product_id": Categorical,
        }
        es.add_dataframe(
            dataframe_name="wishlist_log",
            dataframe=wishlist_df,
            index="id",
            make_index=True,
            time_index="datetime",
            logical_types=logical_types,
        )
        es.add_relationship("sessions", "id", "wishlist_log", "session_id")
        es.add_last_time_indexes()
        sessions = es["sessions"]
        # wishlist df has more recent events for two session ids
        true_sessions_lti[1] = pd.Timestamp("2011-4-9 10:31:30")
        true_sessions_lti[3] = pd.Timestamp("2011-4-10 10:41:00")

        lti_name = sessions.ww.metadata.get("last_time_index")
        assert len(sessions[lti_name]) == 6
        sorted_lti = to_pandas(sessions[lti_name]).sort_index()
        for v1, v2 in zip(sorted_lti, true_sessions_lti):
            assert (pd.isnull(v1) and pd.isnull(v2)) or v1 == v2

    def test_multiple_children_right_missing(self, es, wishlist_df, true_sessions_lti):
        if es.dataframe_type == Library.SPARK:
            pytest.xfail("Cannot make index on a Spark DataFrame")
        # test all instances in left child

        # drop wishlist instance related to id 3 so it's only in log
        wishlist_df.drop(4, inplace=True)
        if es.dataframe_type == Library.DASK:
            wishlist_df = dd.from_pandas(wishlist_df, npartitions=2)
        logical_types = {
            "session_id": Integer,
            "datetime": Datetime,
            "product_id": Categorical,
        }
        es.add_dataframe(
            dataframe_name="wishlist_log",
            dataframe=wishlist_df,
            index="id",
            make_index=True,
            time_index="datetime",
            logical_types=logical_types,
        )
        es.add_relationship("sessions", "id", "wishlist_log", "session_id")
        es.add_last_time_indexes()
        sessions = es["sessions"]

        # now only session id 1 has newer event in wishlist_log
        true_sessions_lti[1] = pd.Timestamp("2011-4-9 10:31:30")

        lti_name = sessions.ww.metadata.get("last_time_index")
        assert len(sessions[lti_name]) == 6
        sorted_lti = to_pandas(sessions[lti_name]).sort_index()
        for v1, v2 in zip(sorted_lti, true_sessions_lti):
            assert (pd.isnull(v1) and pd.isnull(v2)) or v1 == v2

    def test_multiple_children_left_missing(
        self,
        es,
        extra_session_df,
        wishlist_df,
        true_sessions_lti,
    ):
        if es.dataframe_type == Library.SPARK:
            pytest.xfail("Cannot make index on a Spark DataFrame")

        # add row to sessions so not all session instances are in log
        es.replace_dataframe(dataframe_name="sessions", df=extra_session_df)

        # add row to wishlist df so new session instance in in wishlist_log
        row_values = {
            "session_id": [6],
            "datetime": [pd.Timestamp("2011-04-11 11:11:11")],
            "product_id": ["toothpaste"],
        }
        row = pd.DataFrame(row_values, index=pd.RangeIndex(start=7, stop=8))
        df = pd.concat([wishlist_df, row])
        if es.dataframe_type == Library.DASK:
            df = dd.from_pandas(df, npartitions=2)
        logical_types = {
            "session_id": Integer,
            "datetime": Datetime,
            "product_id": Categorical,
        }
        es.add_dataframe(
            dataframe_name="wishlist_log",
            dataframe=df,
            index="id",
            make_index=True,
            time_index="datetime",
            logical_types=logical_types,
        )
        es.add_relationship("sessions", "id", "wishlist_log", "session_id")
        es.add_last_time_indexes()

        # test all instances in right child
        sessions = es["sessions"]

        # now wishlist_log has newer events for 3 session ids
        true_sessions_lti[1] = pd.Timestamp("2011-4-9 10:31:30")
        true_sessions_lti[3] = pd.Timestamp("2011-4-10 10:41:00")
        true_sessions_lti[6] = pd.Timestamp("2011-04-11 11:11:11")

        lti_name = sessions.ww.metadata.get("last_time_index")
        assert len(sessions[lti_name]) == 7
        sorted_lti = to_pandas(sessions[lti_name]).sort_index()
        for v1, v2 in zip(sorted_lti, true_sessions_lti):
            assert (pd.isnull(v1) and pd.isnull(v2)) or v1 == v2

    def test_multiple_children_all_combined(
        self,
        es,
        extra_session_df,
        wishlist_df,
        true_sessions_lti,
    ):
        if es.dataframe_type == Library.SPARK:
            pytest.xfail("Cannot make index on a Spark DataFrame")

        # add row to sessions so not all session instances are in log
        es.replace_dataframe(dataframe_name="sessions", df=extra_session_df)

        # add row to wishlist_log so extra session has child instance
        row_values = {
            "session_id": [6],
            "datetime": [pd.Timestamp("2011-04-11 11:11:11")],
            "product_id": ["toothpaste"],
        }
        row = pd.DataFrame(row_values, index=pd.RangeIndex(start=7, stop=8))
        df = pd.concat([wishlist_df, row])

        # drop instance 4 so wishlist_log does not have session id 3 instance
        df.drop(4, inplace=True)
        if es.dataframe_type == Library.DASK:
            df = dd.from_pandas(df, npartitions=2)
        logical_types = {
            "session_id": Integer,
            "datetime": Datetime,
            "product_id": Categorical,
        }
        es.add_dataframe(
            dataframe_name="wishlist_log",
            dataframe=df,
            index="id",
            make_index=True,
            time_index="datetime",
            logical_types=logical_types,
        )
        es.add_relationship("sessions", "id", "wishlist_log", "session_id")
        es.add_last_time_indexes()

        # test some instances in right, some in left, all when combined
        sessions = es["sessions"]

        # wishlist has newer events for 2 sessions
        true_sessions_lti[1] = pd.Timestamp("2011-4-9 10:31:30")
        true_sessions_lti[6] = pd.Timestamp("2011-04-11 11:11:11")

        lti_name = sessions.ww.metadata.get("last_time_index")
        assert len(sessions[lti_name]) == 7
        sorted_lti = to_pandas(sessions[lti_name]).sort_index()
        for v1, v2 in zip(sorted_lti, true_sessions_lti):
            assert (pd.isnull(v1) and pd.isnull(v2)) or v1 == v2

    def test_multiple_children_both_missing(
        self,
        es,
        extra_session_df,
        wishlist_df,
        true_sessions_lti,
    ):
        if es.dataframe_type == Library.SPARK:
            pytest.xfail("Cannot make index on a Spark DataFrame")
        # test all instances in neither child
        sessions = es["sessions"]

        if es.dataframe_type == Library.DASK:
            wishlist_df = dd.from_pandas(wishlist_df, npartitions=2)

        logical_types = {
            "session_id": Integer,
            "datetime": Datetime,
            "product_id": Categorical,
        }
        # add row to sessions to create session with no events
        es.replace_dataframe(dataframe_name="sessions", df=extra_session_df)

        es.add_dataframe(
            dataframe_name="wishlist_log",
            dataframe=wishlist_df,
            index="id",
            make_index=True,
            time_index="datetime",
            logical_types=logical_types,
        )
        es.add_relationship("sessions", "id", "wishlist_log", "session_id")
        es.add_last_time_indexes()
        sessions = es["sessions"]

        # wishlist has 2 newer events and one is NaT
        true_sessions_lti[1] = pd.Timestamp("2011-4-9 10:31:30")
        true_sessions_lti[3] = pd.Timestamp("2011-4-10 10:41:00")
        true_sessions_lti[6] = pd.NaT

        lti_name = sessions.ww.metadata.get("last_time_index")
        assert len(sessions[lti_name]) == 7
        sorted_lti = to_pandas(sessions[lti_name]).sort_index()
        for v1, v2 in zip(sorted_lti, true_sessions_lti):
            assert (pd.isnull(v1) and pd.isnull(v2)) or v1 == v2

    def test_grandparent(self, es):
        # test sorting by time works correctly across several generations
        log = es["log"]

        # For one user, change a log event to be newer than the user's normal
        # last time index. This event should be from a different session than
        # the current last time index.
        df = to_pandas(log)
        df["datetime"][5] = pd.Timestamp("2011-4-09 10:40:01")
        df = (
            df.set_index("datetime", append=True)
            .sort_index(level=[1, 0], kind="mergesort")
            .reset_index("datetime", drop=False)
        )
        if es.dataframe_type == Library.DASK:
            df = dd.from_pandas(df, npartitions=2)
        if es.dataframe_type == Library.SPARK:
            df = ps.from_pandas(df)
        es.replace_dataframe(dataframe_name="log", df=df)
        es.add_last_time_indexes()
        customers = es["customers"]

        true_customers_lti = pd.Series(
            [
                datetime(2011, 4, 9, 10, 40, 1),
                datetime(2011, 4, 10, 10, 41, 6),
                datetime(2011, 4, 10, 11, 10, 3),
            ],
        )

        lti_name = customers.ww.metadata.get("last_time_index")
        assert len(customers[lti_name]) == 3
        sorted_lti = to_pandas(customers).sort_values("id")[lti_name]
        for v1, v2 in zip(sorted_lti, true_customers_lti):
            assert (pd.isnull(v1) and pd.isnull(v2)) or v1 == v2



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Double

from featuretools.primitives.base import TransformPrimitive


class PercentChange(TransformPrimitive):
    """Determines the percent difference between values in a list.

    Description:
        Given a list of numbers, return the percent difference
        between each subsequent number. Percentages are shown in
        decimal form (not multiplied by 100). Uses pandas' pct_change
        function.

    Args:
        periods (int): Periods to shift for calculating percent change.
            Default is 1.

        fill_method (str): Method for filling gaps in reindexed
            Series. Valid options are `backfill`, `bfill`, `pad`, `ffill`.
            `pad / ffill`: fill gap with last valid observation.
            `backfill / bfill`: fill gap with next valid observation.
            Default is `pad`.

        limit (int): The max number of consecutive NaN values in a gap that
            can be filled. Default is None.

        freq (DateOffset, timedelta, or offset alias string):
            If `freq` is specified, instead of calcualting change between subsequent
            points, PercentChange will calculate change between points with a
            certain interval between their date indices. `freq` defines the
            desired interval. When freq is used, the resulting index will also be
            filled to include any missing dates from the specified interval.

            If the index is not date/datetime and freq is used, it will raise a
            NotImplementedError.

            If freq is None, no changes will be applied. Default is None.

    Examples:
        >>> percent_change = PercentChange()
        >>> percent_change([2, 5, 15, 3, 3, 9, 4.5]).to_list()
        [nan, 1.5, 2.0, -0.8, 0.0, 2.0, -0.5]

        We can control the number of periods to return the percent
            difference between points further from one another.

        >>> percent_change_2 = PercentChange(periods=2)
        >>> percent_change_2([2, 5, 15, 3, 3, 9, 4.5]).to_list()
        [nan, nan, 6.5, -0.4, -0.8, 2.0, 0.5]

        We can control the method used to handle gaps in data.

        >>> percent_change = PercentChange()
        >>> percent_change([2, 4, 8, None, 16, None, 32, None]).to_list()
        [nan, 1.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0]
        >>> percent_change_backfill = PercentChange(fill_method='backfill')
        >>> percent_change_backfill([2, 4, 8, None, 16, None, 32, None]).to_list()
        [nan, 1.0, 1.0, 1.0, 0.0, 1.0, 0.0, nan]

        We can also control the maximum number of NaN values to fill in a gap.

        >>> percent_change = PercentChange()
        >>> percent_change([2, None, None, None, 4]).to_list()
        [nan, 0.0, 0.0, 0.0, 1.0]
        >>> percent_change_limited = PercentChange(limit=2)
        >>> percent_change_limited([2, None, None, None, 4]).to_list()
        [nan, 0.0, 0.0, nan, nan]

        Finally, we can specify a date frequency on which to calculate percent
            change.

        >>> import pandas as pd
        >>> dates = pd.DatetimeIndex(['2018-01-01', '2018-01-02', '2018-01-03', '2018-01-05'])
        >>> x_indexed = pd.Series([1, 2, 3, 4], index=dates)
        >>> percent_change = PercentChange()
        >>> percent_change(x_indexed).to_list()
        [nan, 1.0, 0.5, 0.33333333333333326]
        >>> date_offset = pd.tseries.offsets.DateOffset(days=1)
        >>> percent_change_freq = PercentChange(freq=date_offset)
        >>> percent_change_freq(x_indexed).to_list()
        [nan, 1.0, 0.5, nan]
    """

    name = "percent_change"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(logical_type=Double, semantic_tags={"numeric"})

    def __init__(self, periods=1, fill_method="pad", limit=None, freq=None):
        if fill_method not in ["backfill", "bfill", "pad", "ffill"]:
            raise ValueError("Invalid fill_method")
        self.periods = periods
        self.fill_method = fill_method
        self.limit = limit
        self.freq = freq

    def get_function(self):
        def percent_change(data):
            return data.pct_change(
                self.periods,
                self.fill_method,
                self.limit,
                self.freq,
            )

        return percent_change



import pandas as pd
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Datetime

from featuretools.primitives.base.aggregation_primitive_base import AggregationPrimitive
from featuretools.utils import calculate_trend
from featuretools.utils.gen_utils import Library


class Trend(AggregationPrimitive):
    """Calculates the trend of a column over time.

    Description:
        Given a list of values and a corresponding list of
        datetimes, calculate the slope of the linear trend
        of values.

    Examples:
        >>> from datetime import datetime
        >>> trend = Trend()
        >>> times = [datetime(2010, 1, 1, 11, 45, 0),
        ...          datetime(2010, 1, 1, 11, 55, 15),
        ...          datetime(2010, 1, 1, 11, 57, 30),
        ...          datetime(2010, 1, 1, 11, 12),
        ...          datetime(2010, 1, 1, 11, 12, 15)]
        >>> round(trend([1, 2, 3, 4, 5], times), 3)
        -0.053
    """

    name = "trend"
    input_types = [
        ColumnSchema(semantic_tags={"numeric"}),
        ColumnSchema(logical_type=Datetime, semantic_tags={"time_index"}),
    ]
    return_type = ColumnSchema(semantic_tags={"numeric"})
    description_template = "the linear trend of {} over time"

    def get_function(self, agg_type=Library.PANDAS):
        def pd_trend(y, x):
            return calculate_trend(pd.Series(data=y.values, index=x.values))

        return pd_trend



import hashlib
import json
from functools import lru_cache
from typing import Any, Dict, Tuple

from woodwork.column_schema import ColumnSchema

from featuretools.feature_discovery.type_defs import ANY
from featuretools.primitives.base.primitive_base import PrimitiveBase
from featuretools.primitives.utils import (
    get_all_logical_type_names,
    get_all_primitives,
    serialize_primitive,
)

primitives_map = get_all_primitives()
logical_types_map = get_all_logical_type_names()


def column_schema_to_keys(column_schema: ColumnSchema) -> str:
    """
    Generate a hashing key from a Columns Schema. For example:
    - ColumnSchema(logical_type=Double) -> "Double"
    - ColumnSchema(semantic_tags={"index"}) -> "index"
    - ColumnSchema(logical_type=Double, semantic_tags={"index", "other"}) -> "Double,index,other"

    Args:
        column_schema (ColumnSchema):

    Returns:
        str: hashing key
    """
    logical_type = column_schema.logical_type
    tags = column_schema.semantic_tags
    lt_key = None
    if logical_type:
        lt_key = type(logical_type).__name__

    tags = sorted(tags)
    if len(tags) > 0:
        tag_key = ",".join(tags)
        return f"{lt_key},{tag_key}" if lt_key is not None else tag_key

    elif lt_key is not None:
        return lt_key
    else:
        return ANY


@lru_cache(maxsize=None)
def hash_primitive(primitive: PrimitiveBase) -> Tuple[str, Dict[str, Any]]:
    hash_msg = hashlib.sha256()
    primitive_name = primitive.name
    assert isinstance(primitive_name, str)
    primitive_dict = serialize_primitive(primitive)
    primitive_json = json.dumps(primitive_dict).encode("utf-8")
    hash_msg.update(primitive_json)
    key = hash_msg.hexdigest()
    return (key, primitive_dict)


def get_primitive_return_type(primitive: PrimitiveBase) -> ColumnSchema:
    """
    Get Return type from a primitive

    Args:
        primitive (PrimitiveBase)

    Returns:
        ColumnSchema
    """
    if primitive.return_type:
        return primitive.return_type
    return_type = primitive.input_types[0]
    if isinstance(return_type, list):
        return_type = return_type[0]
    return return_type



import logging
import os

from featuretools.config_init import initialize_logging

logging_env_vars = {
    "FEATURETOOLS_LOG_LEVEL": "debug",
    "FEATURETOOLS_ES_LOG_LEVEL": "critical",
    "FEATURETOOLS_BACKEND_LOG_LEVEL": "error",
}


def test_logging_defaults():
    old_env_vars = {}
    for env_var in logging_env_vars:
        old_env_vars[env_var] = os.environ.get(env_var, None)
        if old_env_vars[env_var] is not None:
            del os.environ[env_var]

    initialize_logging()
    main_logger = logging.getLogger("featuretools")
    assert main_logger.getEffectiveLevel() == logging.INFO
    es_logger = logging.getLogger("featuretools.entityset")
    assert es_logger.getEffectiveLevel() == logging.INFO
    backend_logger = logging.getLogger("featuretools.computation_backend")
    assert backend_logger.getEffectiveLevel() == logging.INFO

    for env_var, value in old_env_vars.items():
        if value is not None:
            os.environ[env_var] = value


def test_logging_set_via_env():
    old_env_vars = {}
    for env_var, value in logging_env_vars.items():
        old_env_vars[env_var] = os.environ.get(env_var, None)
        os.environ[env_var] = value

    initialize_logging()
    main_logger = logging.getLogger("featuretools")
    assert main_logger.getEffectiveLevel() == logging.DEBUG
    es_logger = logging.getLogger("featuretools.entityset")
    assert es_logger.getEffectiveLevel() == logging.CRITICAL
    backend_logger = logging.getLogger("featuretools.computation_backend")
    assert backend_logger.getEffectiveLevel() == logging.ERROR

    for env_var, value in old_env_vars.items():
        if value is not None:
            os.environ[env_var] = value



from woodwork.column_schema import ColumnSchema

from featuretools.primitives.base import TransformPrimitive


class CumMin(TransformPrimitive):
    """Calculates the cumulative minimum.

    Description:
        Given a list of values, return the cumulative min
        (or running min). There is no set window, so the min
        at each point is calculated over all prior values.
        `NaN` values will return `NaN`, but in the window of a
        cumulative caluclation, they're ignored.

    Examples:
        >>> cum_min = CumMin()
        >>> cum_min([1, 2, -3, 4, None, 5]).tolist()
        [1.0, 1.0, -3.0, -3.0, nan, -3.0]
    """

    name = "cum_min"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(semantic_tags={"numeric"})
    uses_full_dataframe = True
    description_template = "the cumulative minimum of {}"

    def get_function(self):
        def cum_min(values):
            return values.cummin()

        return cum_min



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Datetime, Ordinal

from featuretools.primitives.base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class Day(TransformPrimitive):
    """Determines the day of the month from a datetime.

    Examples:
        >>> from datetime import datetime
        >>> dates = [datetime(2019, 3, 1),
        ...          datetime(2019, 3, 3),
        ...          datetime(2019, 3, 31)]
        >>> day = Day()
        >>> day(dates).tolist()
        [1, 3, 31]
    """

    name = "day"
    input_types = [ColumnSchema(logical_type=Datetime)]
    return_type = ColumnSchema(
        logical_type=Ordinal(order=list(range(1, 32))),
        semantic_tags={"category"},
    )
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]
    description_template = "the day of the month of {}"

    def get_function(self):
        def day(vals):
            return vals.dt.day

        return day



import numpy as np
import pandas as pd

from featuretools.primitives import TitleWordCount
from featuretools.tests.primitive_tests.utils import (
    PrimitiveTestBase,
    find_applicable_primitives,
    valid_dfs,
)


class TestTitleWordCount(PrimitiveTestBase):
    primitive = TitleWordCount

    def test_strings(self):
        x = pd.Series(
            [
                "My favorite movie is Jaws.",
                "this is a string",
                "AAA",
                "I bought a Yo-Yo",
            ],
        )
        primitive_func = self.primitive().get_function()
        answers = [2.0, 0.0, 1.0, 2.0]
        np.testing.assert_array_equal(answers, primitive_func(x))

    def test_nan(self):
        x = pd.Series([np.nan, "", "My favorite movie is Jaws."])
        primitive_func = self.primitive().get_function()
        answers = [np.nan, 0.0, 2.0]
        np.testing.assert_array_equal(answers, primitive_func(x))

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        transform.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)



# flake8: noqa
from featuretools.primitives.standard.transform.absolute_diff import AbsoluteDiff
from featuretools.primitives.standard.transform.binary import *
from featuretools.primitives.standard.transform.cumulative import *
from featuretools.primitives.standard.transform.datetime import *
from featuretools.primitives.standard.transform.email import *
from featuretools.primitives.standard.transform.exponential import *
from featuretools.primitives.standard.transform.file_extension import FileExtension
from featuretools.primitives.standard.transform.full_name_to_first_name import (
    FullNameToFirstName,
)
from featuretools.primitives.standard.transform.full_name_to_last_name import (
    FullNameToLastName,
)
from featuretools.primitives.standard.transform.full_name_to_title import (
    FullNameToTitle,
)
from featuretools.primitives.standard.transform.nth_week_of_month import NthWeekOfMonth
from featuretools.primitives.standard.transform.is_in import IsIn
from featuretools.primitives.standard.transform.is_null import IsNull
from featuretools.primitives.standard.transform.latlong import *
from featuretools.primitives.standard.transform.natural_language import *
from featuretools.primitives.standard.transform.not_primitive import Not
from featuretools.primitives.standard.transform.numeric import *
from featuretools.primitives.standard.transform.percent_change import PercentChange
from featuretools.primitives.standard.transform.postal import *
from featuretools.primitives.standard.transform.savgol_filter import SavgolFilter
from featuretools.primitives.standard.transform.time_series import *
from featuretools.primitives.standard.transform.url import *



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Double

from featuretools.primitives.base import AggregationPrimitive


class AverageCountPerUnique(AggregationPrimitive):
    """Determines the average count across all unique value.

    Args:
        skipna (bool): Determines if to use NA/null values.
            Defaults to True to skip NA/null.

    Examples:
        Determine the average count values for all unique items
        in the input
        >>> input = [1, 1, 2, 2, 3, 4, 5, 6, 7, 8]
        >>> avg_count_per_unique = AverageCountPerUnique()
        >>> avg_count_per_unique(input)
        1.25

        Determine the average count values for all unique items
        in the input with nan values ignored
        >>> input = [1, 1, 2, 2, 3, 4, 5, None, 6, 7, 8]
        >>> avg_count_per_unique = AverageCountPerUnique()
        >>> avg_count_per_unique(input)
        1.25

        Determine the average count values for all unique items
        in the input with nan values included
        >>> input = [1, 2, 2, 3, 4, 5, None, 6, 7, 8, 9]
        >>> avg_count_per_unique_skipna_false = AverageCountPerUnique(skipna=False)
        >>> avg_count_per_unique_skipna_false(input)
        1.1
    """

    name = "average_count_per_unique"
    input_types = [ColumnSchema(semantic_tags={"category"})]
    return_type = ColumnSchema(logical_type=Double, semantic_tags={"numeric"})
    default_value = 0

    def __init__(self, skipna=True):
        self.skipna = skipna

    def get_function(self):
        def average_count_per_unique(x):
            return x.value_counts(
                dropna=self.skipna,
            ).mean(skipna=self.skipna)

        return average_count_per_unique



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Double, Integer

from featuretools.primitives.base import AggregationPrimitive


class MaxConsecutiveNegatives(AggregationPrimitive):
    """Determines the maximum number of consecutive negative values in the input

    Args:
        skipna (bool): Ignore any `NaN` values in the input. Default is True.

    Examples:
        >>> max_consecutive_negatives = MaxConsecutiveNegatives()
        >>> max_consecutive_negatives([1.0, -1.4, -2.4, -5.4, 2.9, -4.3])
        3

        `NaN` values can be ignored with the `skipna` parameter

        >>> max_consecutive_negatives_skipna = MaxConsecutiveNegatives(skipna=False)
        >>> max_consecutive_negatives_skipna([1.0, 1.4, -2.4, None, -2.9, -4.3])
        2
    """

    name = "max_consecutive_negatives"
    input_types = [
        [ColumnSchema(logical_type=Integer)],
        [ColumnSchema(logical_type=Double)],
    ]
    return_type = ColumnSchema(logical_type=Integer, semantic_tags={"numeric"})
    stack_on_self = False
    default_value = 0

    def __init__(self, skipna=True):
        self.skipna = skipna

    def get_function(self):
        def max_consecutive_negatives(x):
            if self.skipna:
                x = x.dropna()
            # convert the numeric values to booleans for processing
            x[x.notnull()] = x[x.notnull()].lt(0)
            # find the locations where the value changes from the previous value
            not_equal = x != x.shift()
            # Use cumulative sum to determine where consecutive values occur. When the
            # sum changes, consecutive non-negative values are present, when the cumulative
            # sum remains unchnaged, consecutive negative values are present.
            not_equal_sum = not_equal.cumsum()
            # group the input by the cumulative sum values and use cumulative count
            # to count the number of consecutive values. Add 1 to account for the cumulative
            # sum starting at zero where the first negative occurs
            consecutive = x.groupby(not_equal_sum).cumcount() + 1
            # multiply by the inverted input to keep only the counts that correspond to
            # negative values
            consecutive_neg = consecutive * x
            # return the max of all the consecutive negative values
            return consecutive_neg.max()

        return max_consecutive_negatives



import numpy as np
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import BooleanNullable, LatLong

from featuretools.primitives.base import TransformPrimitive


class IsInGeoBox(TransformPrimitive):
    """Determines if coordinates are inside a box defined by two
    corner coordinate points.

    Description:
        Coordinate values should be specified as (latitude, longitude)
        tuples. This primitive is unable to handle coordinates and boxes
        at the poles, and near +/- 180 degrees latitude.

    Args:
        point1 (tuple(float, float)): The coordinates
            of the first corner of the box. Defaults to (0, 0).
        point2 (tuple(float, float)): The coordinates
            of the diagonal corner of the box. Defaults to (0, 0).

    Example:
        >>> is_in_geobox = IsInGeoBox((40.7128, -74.0060), (42.2436, -71.1677))
        >>> is_in_geobox([(41.034, -72.254), (39.125, -87.345)]).tolist()
        [True, False]
    """

    name = "is_in_geobox"
    input_types = [ColumnSchema(logical_type=LatLong)]
    return_type = ColumnSchema(logical_type=BooleanNullable)

    def __init__(self, point1=(0, 0), point2=(0, 0)):
        self.point1 = point1
        self.point2 = point2
        self.lats = np.sort(np.array([point1[0], point2[0]]))
        self.lons = np.sort(np.array([point1[1], point2[1]]))

    def get_function(self):
        def geobox(latlongs):
            transposed = np.transpose(np.array(latlongs.tolist()))
            lats = (self.lats[0] <= transposed[0]) & (self.lats[1] >= transposed[0])
            longs = (self.lons[0] <= transposed[1]) & (self.lons[1] >= transposed[1])
            return lats & longs

        return geobox



def convert_to_nth(integer):
    string_nth = str(integer)
    end_int = integer % 10
    if end_int == 1 and integer % 100 != 11:
        return str(integer) + "st"
    elif end_int == 2 and integer % 100 != 12:
        return str(string_nth) + "nd"
    elif end_int == 3 and integer % 100 != 13:
        return str(string_nth) + "rd"
    else:
        return str(string_nth) + "th"



import os

import pytest

from featuretools import __version__
from featuretools.utils import (
    get_featuretools_root,
    get_installed_packages,
    get_sys_info,
    show_info,
)


@pytest.fixture
def this_dir():
    return os.path.dirname(os.path.abspath(__file__))


def test_show_info(capsys):
    show_info()
    captured = capsys.readouterr()
    assert "Featuretools version" in captured.out
    assert "Featuretools installation directory:" in captured.out
    assert __version__ in captured.out
    assert "SYSTEM INFO" in captured.out


def test_sys_info():
    sys_info = get_sys_info()
    info_keys = [
        "python",
        "python-bits",
        "OS",
        "OS-release",
        "machine",
        "processor",
        "byteorder",
        "LC_ALL",
        "LANG",
        "LOCALE",
    ]
    found_keys = [k for k, _ in sys_info]
    assert set(info_keys).issubset(found_keys)


def test_installed_packages():
    installed_packages = get_installed_packages()
    # Per PEP 426, package names are case insensitive
    # Underscore and hyphen are equivalent
    installed_set = {
        name.lower().replace("-", "_") for name in installed_packages.keys()
    }
    requirements = [
        "pandas",
        "numpy",
        "tqdm",
        "cloudpickle",
        "psutil",
    ]
    assert set(requirements).issubset(installed_set)


def test_get_featuretools_root(this_dir):
    root = os.path.abspath(os.path.join(this_dir, "..", ".."))
    assert get_featuretools_root() == root



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Datetime, Ordinal

from featuretools.primitives.base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class DayOfYear(TransformPrimitive):
    """Determines the ordinal day of the year from the given datetime

    Description:
        For a list of dates, return the ordinal day of the year
        from the given datetime.

    Examples:
        >>> from datetime import datetime
        >>> dates = [datetime(2019, 1, 1),
        ...          datetime(2020, 12, 31),
        ...          datetime(2020, 2, 28)]
        >>> dayOfYear = DayOfYear()
        >>> dayOfYear(dates).tolist()
        [1, 366, 59]
    """

    name = "day_of_year"
    input_types = [ColumnSchema(logical_type=Datetime)]
    return_type = ColumnSchema(
        logical_type=Ordinal(order=list(range(1, 367))),
        semantic_tags={"category"},
    )
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]
    description_template = "the day of year from {}"

    def get_function(self):
        def dayOfYear(vals):
            return vals.dt.dayofyear

        return dayOfYear



import pandas as pd
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import BooleanNullable

from featuretools.primitives.base.transform_primitive_base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class Equal(TransformPrimitive):
    """Determines if values in one list are equal to another list.

    Description:
        Given a list of values X and a list of values Y, determine
        whether each value in X is equal to each corresponding value
        in Y.

    Examples:
        >>> equal = Equal()
        >>> equal([2, 1, 2], [1, 2, 2]).tolist()
        [False, False, True]
    """

    name = "equal"
    input_types = [ColumnSchema(), ColumnSchema()]
    return_type = ColumnSchema(logical_type=BooleanNullable)
    commutative = True
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]
    description_template = "whether {} equals {}"

    def get_function(self):
        def equal(x_vals, y_vals):
            if isinstance(x_vals.dtype, pd.CategoricalDtype) and isinstance(
                y_vals.dtype,
                pd.CategoricalDtype,
            ):
                categories = set(x_vals.cat.categories).union(
                    set(y_vals.cat.categories),
                )
                x_vals = x_vals.cat.add_categories(
                    categories.difference(set(x_vals.cat.categories)),
                )
                y_vals = y_vals.cat.add_categories(
                    categories.difference(set(y_vals.cat.categories)),
                )
            return x_vals.eq(y_vals)

        return equal

    def generate_name(self, base_feature_names):
        return "%s = %s" % (base_feature_names[0], base_feature_names[1])



# flake8: noqa
from featuretools.tests.testing_utils.cluster import (
    MockClient,
    mock_cluster,
    get_mock_client_cluster,
)
from featuretools.tests.testing_utils.es_utils import get_df_tags, to_pandas
from featuretools.tests.testing_utils.features import (
    feature_with_name,
    number_of_features_with_name_like,
    backward_path,
    forward_path,
    check_rename,
    check_names,
)
from featuretools.tests.testing_utils.mock_ds import make_ecommerce_entityset



import pandas as pd
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Datetime

from featuretools.primitives.base import TransformPrimitive
from featuretools.primitives.standard.transform.datetime.utils import HolidayUtil


class DistanceToHoliday(TransformPrimitive):
    """Computes the number of days before or after a given holiday.

    Description:
        For a list of dates, return the distance from the nearest
        occurrence of a chosen holiday. The distance is returned in
        days. If the closest occurrence is prior to the date given,
        return a negative number.

        If a date is missing, return `NaN`.

        Currently only works with dates between 1950 and 2100.

    Args:
        holiday (str): Name of the holiday. Defaults to New Year's Day.

        country (str): Specifies which country's calendar to use for the
            given holiday. Default is `US`.

    Examples:
        >>> from datetime import datetime
        >>> distance_to_holiday = DistanceToHoliday("New Year's Day")
        >>> dates = [datetime(2010, 1, 1),
        ...          datetime(2012, 5, 31),
        ...          datetime(2017, 7, 31),
        ...          datetime(2020, 12, 31)]
        >>> distance_to_holiday(dates).tolist()
        [0, -151, 154, 1]

        We can also control the country in which we're searching for
            a holiday.

        >>> distance_to_holiday = DistanceToHoliday("Victoria Day", country='Canada')
        >>> dates = [datetime(2010, 1, 1),
        ...          datetime(2012, 5, 31),
        ...          datetime(2017, 7, 31),
        ...          datetime(2020, 12, 31)]
        >>> distance_to_holiday(dates).tolist()
        [143, -10, -70, 144]
    """

    name = "distance_to_holiday"
    input_types = [ColumnSchema(logical_type=Datetime)]
    return_type = ColumnSchema(semantic_tags={"numeric"})
    default_value = 0

    def __init__(self, holiday="New Year's Day", country="US"):
        self.country = country
        self.holiday = holiday
        self.holidayUtil = HolidayUtil(country)

        available_holidays = list(set(self.holidayUtil.federal_holidays.values()))
        if self.holiday not in available_holidays:
            error = "must be one of the available holidays:\n%s" % available_holidays
            raise ValueError(error)

    def get_function(self):
        def distance_to_holiday(x):
            holiday_df = self.holidayUtil.to_df()
            holiday_df = holiday_df[holiday_df.names == self.holiday]

            df = pd.DataFrame({"date": x})
            df["x_index"] = df.index  # store original index as a column
            df = df.dropna()
            df = df.sort_values("date")
            df["date"] = df["date"].dt.date.astype("datetime64[ns]")

            matches = pd.merge_asof(
                df,
                holiday_df,
                left_on="date",
                right_on="holiday_date",
                direction="nearest",
                tolerance=pd.Timedelta("365d"),
            )
            matches = matches.set_index("x_index")
            matches["days_diff"] = (matches.holiday_date - matches.date).dt.days

            return matches.days_diff.reindex_like(x)

        return distance_to_holiday
